import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

from pysamosa.common_types import (
    RetrackerBaseType,
    ModelSettings,
    WaveformSettings,
    L1bSourceType,
)
from pysamosa.montecarlo_simulator import (
    MonteCarloSimulator,
    plot_cost_func_vs_swh,
    plot_swh_epoch_scatter,
    CostFunctionType,
    default_cost_functions,
    cost_functions,
)
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset


retracker_basetype, settings_preset = (
    # RetrackerBaseType.SAM, SettingsPreset.NONE,
    RetrackerBaseType.SAM,
    SettingsPreset.CORALv1,
    # RetrackerBaseType.SAMPLUS, SettingsPreset.NONE,
    # RetrackerBaseType.SAM, SettingsPreset.SAM_FLO,
    # RetrackerBaseType.SAMPLUS, SettingsPreset.SAMPLUS_FLO,
    # RetrackerBaseType.SAMPLUSPLUS, SettingsPreset.NONE,
)
rp_sets, retrack_sets, fitting_sets, wf_sets, sensor_sets = get_default_base_settings(
    retracker_basetype=retracker_basetype,
    settings_preset=settings_preset,
    l1b_src_type=L1bSourceType.EUM_S3,
)

model_sets = ModelSettings.get_default_sets(
    st=sensor_sets.sensor_type,
    retracker_basetype=retracker_basetype,
    wf_sets=wf_sets,
)

wf_sets_model = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3)


@pytest.fixture
def monte_sim():
    return MonteCarloSimulator(
        rbt=retracker_basetype,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        model_sets=model_sets,
        wf_sets_retracker=wf_sets,
        wf_sets_model=wf_sets_model,
    )


def test_monte_carlo_sim_single_core(monte_sim):
    fit_res_list = []
    n_realisations = 2
    swhs = np.arange(1.0, 8.0, 1)
    for swh in swhs:
        fit_res_list.append(monte_sim(swh=swh, n_realisations=n_realisations))

    df = pd.concat(fit_res_list, ignore_index=True)

    assert df.shape[0] == (swhs.size * n_realisations)


def test_monte_carlo_sim_w_set_param(monte_sim):
    fit_res_list = []
    n_realisations = 2
    swhs = np.arange(1.0, 15.0, 1)
    for swh in swhs:
        fit_res_list.append(
            monte_sim(
                swh=swh,
                n_realisations=n_realisations,
                add_set_name="retrack_sets.leading_edge_weight_factor",
            )
        )

    df = pd.concat(fit_res_list, ignore_index=True)

    assert df.shape[0] == (swhs.size * n_realisations)
    assert "leading_edge_weight_factor" in df.columns


def test_sim_swh_rmse(monte_sim):
    df = monte_sim.multi_proc(swh=np.arange(0.5, 15, 2), n_realisations=1)

    var_types_list = ["swh", "epoch_ns", "Pu"]
    fig, axs = plt.subplots(len(var_types_list), len(default_cost_functions))
    ax_it = iter(axs.ravel())

    for var_type in ["swh", "epoch_ns", "Pu"]:
        for cfunc in default_cost_functions:
            plot_cost_func_vs_swh(
                df, var_type=var_type, cost_func=cfunc, ax=next(ax_it)
            )

    fig.show()


def test_sim_swh_rmse_different_le_weight_factor(monte_sim):
    param_name = "retrack_sets.leading_edge_weight_factor"
    param_name_short = param_name.split(".")[-1]
    params_vals = [1.0, 2.0, 4.0, 8.0]
    add_interference = True
    n_realisations = 1

    # run Monte-Carlo sim with different varying parameters
    # df = pd.DataFrame()
    df_list = []
    for v in params_vals:
        monte_sim.retrack_sets.leading_edge_weight_factor = v
        df_list.append(
            monte_sim.multi_proc(
                swh=np.arange(1, 15, 2),
                n_realisations=n_realisations,
                add_set_name=param_name,
                # max_workers=1,
                # add_interference=add_interference
            )
        )

    df = pd.concat(df_list)

    # prepare plots
    var_types_list = ["swh", "epoch_ns", "Pu"]
    fig, axs = plt.subplots(len(var_types_list), len(default_cost_functions))
    ax_it = iter(axs.ravel())

    # plotting
    for var_type in var_types_list:
        for cfunc in default_cost_functions:
            ax = next(ax_it)
            for v in params_vals:
                sub_df = df.query(f"{param_name_short} == {v}")

                plot_cost_func_vs_swh(
                    sub_df,
                    var_type=var_type,
                    cost_func=cfunc,
                    ax=ax,
                    ax_kwargs={"label": f"{param_name_short}: {v}"},
                )

    fig.suptitle(
        f"{retracker_basetype.value.upper()}, Preset={settings_preset.value.upper()}, N={n_realisations}, state_interference={str(add_interference)}",
        fontsize=6,
    )
    fig.show()


def test_sim_swh_rmse_oversampling():
    param_name = "wf_sets_retracker.internal_oversampling_factor"
    param_name_short = param_name.split(".")[-1]
    params_vals = [1.0, 4.0]
    # params_vals = [4.0]
    # add_thermal_speckle_noise = True
    add_thermal_speckle_noise = False
    add_interference = False
    # n_realisations = 20
    n_realisations = 1
    swh_step = 0.25
    # swh_step = 0.05
    swh_vals = np.arange(-0.125, 20, swh_step)

    # run Monte-Carlo sim with different varying parameters
    df_list = []
    for v in params_vals:
        wf_sets_retracker = WaveformSettings.get_default_src_type(
            L1bSourceType.EUM_S3, internal_oversampling_factor=v
        )
        wf_sets_model = WaveformSettings.get_default_src_type(
            L1bSourceType.EUM_S3, internal_oversampling_factor=1.0
        )
        monte_sim = MonteCarloSimulator(
            rbt=retracker_basetype,
            retrack_sets=retrack_sets,
            fitting_sets=fitting_sets,
            model_sets=model_sets,
            wf_sets_retracker=wf_sets_retracker,
            wf_sets_model=wf_sets_model,
        )

        df_list.append(
            monte_sim.multi_proc(
                swh=swh_vals,
                n_realisations=n_realisations,
                add_set_name=param_name,
                # max_workers=1,
                add_thermal_speckle_noise=add_thermal_speckle_noise,
                add_interference=add_interference,
            )
        )

    df = pd.concat(df_list)

    # prepare plots
    # var_types_list = ['swh', 'epoch_ns', 'Pu']
    var_types_list = ["swh"]
    fig, axs = plt.subplots(
        len(var_types_list) + 1, len(default_cost_functions), dpi=300
    )
    ax_it = iter(axs.ravel())

    # plotting cost funcs for
    for var_type in var_types_list:
        for cfunc in default_cost_functions:
            ax = next(ax_it)
            for v in params_vals:
                sub_df = df.query(f"{param_name_short} == {v}")

                plot_cost_func_vs_swh(
                    sub_df,
                    var_type=var_type,
                    cost_func=cfunc,
                    ax=ax,
                    ax_kwargs={"label": f"{param_name_short}: {v}"},
                )

    # SWH STD plot
    ax = next(ax_it)
    for v in params_vals:
        sub_df = df.query(f"{param_name_short} == {v}")

        plot_cost_func_vs_swh(
            sub_df,
            var_type="swh",
            cost_func=CostFunctionType.SDD,
            ax=ax,
            ax_kwargs={"label": f"{param_name_short}: {v}"},
        )

    # SWH corrections plot
    swh_ranges = np.arange(-0.5, 20 + swh_step, swh_step)
    df_1 = df.query(f"{param_name_short} == {1.0}")
    df_4 = df.query(f"{param_name_short} == {4.0}")

    ax = next(ax_it)
    df_median_bias_vs_swh_exp_1 = df_1.groupby(
        pd.cut(df_1["swh"], swh_ranges, right=False)
    ).apply(
        lambda g: cost_functions[CostFunctionType.MEDIAN_BIAS](
            g["swh"], g["swh_expected"]
        )
    )
    df_median_bias_vs_swh_exp_4 = df_4.groupby(
        pd.cut(df_4["swh"], swh_ranges, right=False)
    ).apply(
        lambda g: cost_functions[CostFunctionType.MEDIAN_BIAS](
            g["swh"], g["swh_expected"]
        )
    )
    df_diff = df_median_bias_vs_swh_exp_4 - df_median_bias_vs_swh_exp_1
    df_diff *= -1  # the correction shall latter be added to the estimated SWH

    # modulate corr below boundary where there are  no values
    ind_first_valid_val = (~np.isnan(df_diff[0:10])).values.nonzero()[0][0]
    m = df_diff[ind_first_valid_val] / ind_first_valid_val
    df_diff[:ind_first_valid_val] = np.arange(ind_first_valid_val) * m

    fontsize_ticks = 7
    fontsize_xylabel = 6

    # replace ranges with center bins (for plotting)
    df_diff.index = [i.mid for i in df_diff.index]
    df_diff = df_diff.fillna(value=0.0)
    _ax = df_diff.plot(ax=ax, marker="o", markersize=3)

    _ax.grid()
    _ax.set_xticks(np.arange(0, int(df_diff.index.values[-1]) + 1))
    _ax.tick_params(labelsize=fontsize_ticks)
    _ax.set_xlabel("estimated SWH [m]", fontsize=fontsize_xylabel)
    _ax.set_ylabel("SWH correction [m]", fontsize=fontsize_xylabel)

    fig.suptitle(
        f"{retracker_basetype.value.upper()}, "
        f"Preset={settings_preset.value.upper()}, "
        f"N={n_realisations}, "
        f"state_thermal_speckle_noise={str(add_thermal_speckle_noise)}, "
        f"state_interference={str(add_interference)}",
        fontsize=6,
    )

    plt.tight_layout()
    plt.subplots_adjust(wspace=0.2, hspace=0.2, top=0.95)
    fig.show()

    # save corr LUT to .pickle
    # dest_pickle_swh_corr = Path.cwd().parent.parent / 'scripts' / f'{settings_preset.value.lower()}_swh_corr_lut.pickle'
    # pickle.dump({'lut_swh_corr': np.asarray(df_diff), 'swh_ranges': np.asarray(swh_ranges)}, open(dest_pickle_swh_corr, 'wb'))
    # swh_ranges_center = swh_ranges[:-1] + swh_step
    # plot_corr_lut_table(swh_ranges_center=swh_ranges_center, corr_lut=np.asarray(df_diff), title=dest_pickle_swh_corr.name)
    # print(f'stored SWH_corr_LUT to {dest_pickle_swh_corr}')


def test_swh_epoch_scatter(monte_sim):
    df = monte_sim.multi_proc(swh=np.arange(2.0), n_realisations=5)

    fig, ax = plt.subplots()
    plot_swh_epoch_scatter(df, ax=ax)
    fig.show()

    print("bla")



from pathlib import Path
import netCDF4

import numpy as np
import pandas as pd
from bokeh.io import show
from bokeh.models import HoverTool
from bokeh.models.glyphs import Circle
from bokeh.plotting import figure, ColumnDataSource
from bokeh.tile_providers import get_provider, Vendors

from pysamosa.data_access import _read_dataset_vars_from_ds

data_vars_eumetsat_s6 = {
    "time": "time",
    "lat_rad": "latitude",
    "lon_rad": "longitude",
}

# data_vars_eumetsat_s6 = {
#     'l1a': {
#         'time': 'time',
#         'lat_rad': 'latitude',
#         'lon_rad': 'longitude',
#     },
#     'l1b': {
#         'time': 'time',
#         'lat_rad': 'latitude',
#         'lon_rad': 'longitude',
#     },
#     'l2': {
#         'time': 'time',
#         'lat_rad': 'latitude',
#         'lon_rad': 'longitude',
#     },
# }

data_vars_eumetsat_cs = {
    "l1a": {
        "time": "time_20_ku",
        "lat_rad": "lat_20_ku",
        "lon_rad": "lon_20_ku",
    },
    "l1b": {
        "time": "time_20_ku",
        "lat_rad": "lat_20_ku",
        "lon_rad": "lon_20_ku",
    },
    "l2": {
        "time": "time_20_ku",
        "lat_rad": "lat_20_ku",
        "lon_rad": "lon_20_ku",
    },
}

data_vars_eumetsat_s3 = {
    "l1a": {
        "time": "time_l1a_echo_sar_ku",
        "lat_rad": "lat_l1a_echo_sar_ku",
        "lon_rad": "lon_l1a_echo_sar_ku",
    },
    "l1b": {
        "time": "time_l1b_echo_sar_ku",
        "lat_rad": "lat_l1b_echo_sar_ku",
        "lon_rad": "lon_l1b_echo_sar_ku",
    },
    "l2": {
        "time": "time_l1bs_echo_sar_ku",
        "lat_rad": "lat_l1bs_echo_sar_ku",
        "lon_rad": "lon_l1bs_echo_sar_ku",
    },
}

nc_filename = Path(
    "/nfs/DGFI145/C/work_flo/pysamosa_lfsdata/s3/l1b/S3A_SR_1_SRA____20180414T041038_20180414T050107_20190706T235251_3029_030_090______MR1_R_NT_004.nc"
)

level_str = (
    str(nc_filename.name).split("_")[2]
    if "meas" not in str(nc_filename)
    else str(nc_filename.parent.name).split("_")[2]
).lower()
level_str = level_str.replace("_1_", "_1b_")

if level_str == "":
    level_str = "1a"

is_l1a = "_1a_" in str(nc_filename).lower()
is_l2 = "_2_" in str(nc_filename).lower()
grps = [g for g in list(netCDF4.Dataset(nc_filename).groups) if "data_20" in g]
grp = f"{grps[0]}/ku" if len(grps) > 0 else ""

if "S6" in str(nc_filename):
    data_vars = data_vars_eumetsat_s6
elif "CS" in str(nc_filename):
    data_vars = data_vars_eumetsat_cs
else:
    data_vars = data_vars_eumetsat_s3["l2"] if is_l2 else data_vars_eumetsat_s3["l1b"]

ds = _read_dataset_vars_from_ds(
    nc_filename=nc_filename, data_var_names=data_vars, group=grp
)
# nc_filename = LFSDATA_DIR / 'eumetsat/l1bs/S3A_SR_1_SRA_BS_20171210T145735_20171210T154803_20190705T205005_3028_025_239______MR1_R_NT_004.nc'
# ds = _read_dataset_vars_from_ds(nc_filename=nc_filename, data_var_names=data_vars_eumetsat['l1b'])

df = pd.DataFrame(
    {
        k: v
        for k, v in ds.items()
        if k in ["time", "lat_rad", "lon_rad", "alt_m", "dist2coast", "record_inds"]
    }
)
df["lat"] = np.degrees(df["lat_rad"])
df["lon"] = np.degrees(df["lon_rad"])

# reduce data
if is_l1a:
    df = df[:: 140 // 2]
elif is_l2:
    # df = df[::20]
    df = df[::]
else:
    pass

# Define function to switch from lat/lon to mercator coordinates


def x_coord(lat, lon):
    r_major = 6378137.000
    x = r_major * np.radians(lon)
    scale = x / lon
    y = (
        180.0
        / np.pi
        * np.log(np.tan(np.pi / 4.0 + lat * (np.pi / 180.0) / 2.0))
        * scale
    )
    return (x, y)


# Define coord as tuple (lat,long)
df["coordinates"] = list(zip(df["lat"], df["lon"]))

# Create mercator column in our df
df["mercator"] = [x_coord(lat, lon) for lat, lon in df["coordinates"]]

# Split that column out into two separate columns - mercator_x and mercator_y
df[["mercator_x", "mercator_y"]] = df["mercator"].apply(pd.Series)

# Plotting
# palette = PRGn[11]  # Choose palette

# Define color mapper - which column will define the colour of the data points
# color_mapper = linear_cmap(field_name = 'AveragePrice', palette = palette, low = df['AveragePrice'].min(), high = df['AveragePrice'].max())

# Set tooltips - these appear when we hover over a data point in our map,
# very nifty and very useful
tooltips = [
    ("latitude", "@lat"),
    ("longitude", "@lon"),
    ("dist2coast", "@dist2coast"),
    ("record_ind", "@record_inds"),
]

# Create the figure
p = figure(
    title=f"{nc_filename.parent.name}/{nc_filename.stem}",
    # plot_width=1200, plot_height=900,
    sizing_mode="scale_width",
    aspect_ratio=2.0,
    x_axis_type="mercator",
    y_axis_type="mercator",
    x_axis_label="longitude",
    y_axis_label="latitude",
    # tooltips=tooltips,
    tools="pan,wheel_zoom,box_zoom,save",
    toolbar_location="above",
    output_backend="webgl",
)

# Add map tile
p.add_tile(get_provider(Vendors.ESRI_IMAGERY))

# Tell Bokeh to use df as the source of the data
df = df.drop(columns=["time"])
source = ColumnDataSource(data=df)

p.circle(
    x="mercator_x",
    y="mercator_y",
    # color=color_mapper,
    color="red",
    fill_alpha=0.9,
    source=source,
    # size=30,
    # radius=dist_lat_merc,
    radius=500,
)

# add Crete transponder
coords_crete_tr_deg = (35.3379, 23.7795)
coords_crete_tr = x_coord(*coords_crete_tr_deg)

source_tr = ColumnDataSource(
    {
        "x": [coords_crete_tr[0]],
        "y": [coords_crete_tr[1]],
        "radius": [700]
        # 'lat': coords_crete_tr_deg[0],
        # 'lon': coords_crete_tr_deg[1],
    }
)

glyph_circle_tr = Circle(
    x="x",
    y="y",
    radius="radius",
    radius_units="data",
    fill_color="yellow",
    fill_alpha=0.9,
    line_color="black",
)

p.add_glyph(source_tr, glyph_circle_tr)

# Add a hover tool referring to the formatted columns
hover = HoverTool(tooltips=tooltips)
hover.point_policy = "follow_mouse"
# hover.point_policy = 'snap_to_data'
# hover.anchor = 'top_left'

# Add the hover tool to the graph
p.add_tools(hover)

# Display in notebook
# output_notebook()
# output_file('track_browser.html')
# output_file()

# Show map
show(p)



import matplotlib.pyplot as plt

from pysamosa import retracker
from pysamosa import simple_logger
from pysamosa.common_types import (
    RetrackerBaseType,
    ModelSettings,
    WaveformSettings,
    L1bSourceType,
    SensorSettings,
    SENSOR_SETS_DEFAULT_S3,
)
from pysamosa.data_access import get_model_param_obj_from_l1b_data
from pysamosa.l1b_simulator import L1bSimulator
from pysamosa.settings_manager import SettingsPreset, get_default_base_settings
from tests.helpers import plot_retrack_result

l1b_src_type, retracker_basetype, settings_preset = (
    # L1bSourceType.EUM_S3, RetrackerBaseType.SAM, SettingsPreset.NONE,
    L1bSourceType.EUM_S3,
    RetrackerBaseType.SAM,
    SettingsPreset.CORALv1,
    # L1bSourceType.EUM_S3, RetrackerBaseType.SAMPLUS, SettingsPreset.NONE,
    # L1bSourceType.EUM_S3, RetrackerBaseType.SAM, SettingsPreset.SAM_FLO,
    # L1bSourceType.EUM_S3, RetrackerBaseType.SAMPLUS, SettingsPreset.SAMPLUS_FLO,
    # L1bSourceType.EUM_S3, RetrackerBaseType.SAMPLUSPLUS, SettingsPreset.NONE,
)
rp_sets, retrack_sets, fitting_sets, wf_sets, sensor_sets = get_default_base_settings(
    retracker_basetype=retracker_basetype,
    settings_preset=settings_preset,
    l1b_src_type=l1b_src_type,
)

# retrack_sets.leading_edge_weight_factor = 1.0


def test_retrack_l1bsim_w_interference():
    n_realisations = 5
    swh = 2.5

    model_sets = ModelSettings.get_default_sets(
        st=sensor_sets.sensor_type,
        retracker_basetype=retracker_basetype,
        n_effective_looks=0,
        wf_sets=wf_sets,
    )

    l1b_sim = L1bSimulator(
        model_sets=model_sets,
        swh=swh,
        Pu=1.0,
        sensor_sets=SENSOR_SETS_DEFAULT_S3,
        add_thermal_speckle_noise=True,
        add_interference=True,
        wf_sets=WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3),
    )
    l1b_sim_it = iter(l1b_sim)

    # run L1B simulations and retracking
    for i in range(n_realisations):
        fig, ax = plt.subplots()
        l1b_data_single = next(l1b_sim_it)
        model_params = get_model_param_obj_from_l1b_data(l1b_data_single, ind=0)

        simple_logger.set_root_logger()

        sr = retracker.SamosaRetracker(
            retracker_basetype=retracker_basetype,
            retrack_sets=retrack_sets,
            fitting_sets=fitting_sets,
            wf_sets=wf_sets,
            sensor_sets=sensor_sets,
        )

        # start fitting
        res_fit = sr.fit_wf(l1b_data_single=l1b_data_single, model_params=model_params)

        # plot fitted curve
        l2_data_single_sim = l1b_sim.get_l2_data_single()
        plot_retrack_result(
            l1b_data_single,
            l2_data_single_sim,
            res_fit,
            model_params=model_params,
            model_sets=sr.model_sets,
            retrack_sets=retrack_sets,
            wf_sets=wf_sets,
            sensor_sets=sensor_sets,
            ax=ax,
            show_l2_model=False,
        )

        fig.suptitle(
            f'Simulated L1B-waveform, SWH={l2_data_single_sim["swh"]}m', fontsize=8
        )
        fig.show()


def test_retrack_l1bsim_wo_interference():
    n_realisations = 1
    swh = 2.5
    # swh = 17

    model_sets = ModelSettings.get_default_sets(
        st=sensor_sets.sensor_type,
        retracker_basetype=retracker_basetype,
        n_effective_looks=0,
        wf_sets=wf_sets,
    )

    l1b_sim = L1bSimulator(
        model_sets=model_sets,
        swh=swh,
        Pu=1.0,
        sensor_sets=SensorSettings(),
        add_thermal_speckle_noise=True,
        # add_thermal_speckle_noise=False,
        add_interference=False,
        wf_sets=WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3),
    )
    l1b_sim_it = iter(l1b_sim)

    # run L1B simulations and retracking
    for i in range(n_realisations):
        fig, ax = plt.subplots()
        l1b_data_single = next(l1b_sim_it)
        model_params = get_model_param_obj_from_l1b_data(l1b_data_single, ind=0)

        simple_logger.set_root_logger()

        sr = retracker.SamosaRetracker(
            retracker_basetype=retracker_basetype,
            retrack_sets=retrack_sets,
            fitting_sets=fitting_sets,
            wf_sets=wf_sets,
            sensor_sets=SENSOR_SETS_DEFAULT_S3,
        )

        # start fitting
        res_fit = sr.fit_wf(l1b_data_single=l1b_data_single, model_params=model_params)

        # plot fitted curve
        l2_data_single_sim = l1b_sim.get_l2_data_single()
        plot_retrack_result(
            l1b_data_single,
            l2_data_single_sim,
            res_fit,
            model_params=model_params,
            model_sets=sr.model_sets,
            retrack_sets=retrack_sets,
            wf_sets=wf_sets,
            sensor_sets=SensorSettings(),
            ax=ax,
            show_l2_model=False,
        )

        fig.subplots_adjust(top=0.7)
        fig.suptitle(
            f'Simulated L1B-waveform, SWH={l2_data_single_sim["swh"]}m', fontsize=8
        )
        fig.show()



import numpy as np
from scipy import signal

from pysamosa.retracker_helpers import get_inds_sliding_window


class SwhQualFlagEstimator:
    def __init__(self, swh: np.ndarray, ssh: np.ndarray):
        self.swh = np.asarray(swh)
        self.ssh = np.asarray(ssh)

        # init qual_flag with nan values (1.0)
        self.swh_qual = np.full(swh.shape, np.nan)
        self.swh_qual[np.isnan(self.swh) | np.isnan(self.ssh)] = 1.0

    def estimate_swh_qual_flag(
        self, n_window_size: int = 100, sigma_boundary: int = 1.5
    ):
        """Estimates a new swh qual flag based on a smart algortihm

        :param swh: significant wave height
        :param ssh: sea surface height
        :param n_window_size: window size over which the window is shifted over
        :return: the new q
        """

        for i, (inds, center_sample) in enumerate(
            get_inds_sliding_window(self.swh.shape[0], n=n_window_size)
        ):
            ssh_seg = self.ssh[inds]
            self.swh[inds]

            non_nans = (~np.isnan(ssh_seg)).nonzero()[0]
            if center_sample in non_nans:
                ssh_detrended = signal.detrend(ssh_seg[non_nans])
                sigma = np.std(ssh_detrended)

                # mark good inds in qual vec
                center_sample_wo_nans = int((non_nans == center_sample).nonzero()[0])
                if np.abs(ssh_detrended[center_sample_wo_nans]) < (
                    sigma_boundary * sigma
                ):
                    self.swh_qual[i] = 0.0
                else:
                    self.swh_qual[i] = 1.0

        return self.swh_qual



import os
import logging
import datetime


class DeltaTimeFormatter(logging.Formatter):
    def format(self, record):
        now = datetime.datetime.now()
        self.last_timestamp = (
            now if "last_timestamp" not in dir(self) else self.last_timestamp
        )
        delta_last = now - self.last_timestamp
        self.last_timestamp = now
        hours, _ = divmod(delta_last.seconds, 3600)
        minutes, seconds = divmod(delta_last.seconds, 60)
        record.delta_last = f"{hours:02d}:{minutes // 60:02d}:{seconds:02d}"

        duration = datetime.datetime.utcfromtimestamp(record.relativeCreated / 1000)
        record.delta_start = duration.strftime("%H:%M:%S")

        return super().format(record)


format_str = "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s (to_start: %(delta_start)s, to_last: %(delta_last)s)"


def set_root_logger(log_level=logging.DEBUG):
    for lh in logging.root.handlers:
        logging.root.removeHandler(lh)

    logging.root.setLevel(log_level)

    # add stream handler
    sh = logging.StreamHandler()
    sh.setLevel(log_level)
    sh.setFormatter(DeltaTimeFormatter(format_str))
    logging.root.addHandler(sh)

    logging.getLogger("matplotlib.font_manager").disabled = True


def generate_and_add_file_logger(log_filename):
    # add file handler
    # create new logfile
    os.umask(0)
    os.makedirs(os.path.dirname(log_filename), exist_ok=True)
    if os.path.exists(log_filename):
        os.remove(log_filename)

    fh = logging.FileHandler(log_filename)
    fh.setFormatter(DeltaTimeFormatter(format_str))
    fh.setLevel(logging.DEBUG)

    logging.root.addHandler(fh)

    return fh



import logging

import numpy as np
from scipy.interpolate import Akima1DInterpolator
from scipy.optimize import least_squares

from pysamosa.common_types import (
    ModelSettings,
    ModelParameter,
    RetrackerSettings,
    FittingSettings,
    WaveformSettings,
    FittingParameters,
)
from pysamosa.common_types import RetrackerBaseType, SensorSettings
from pysamosa.leading_edge_detector import detect_leading_edge
from pysamosa.model import SamosaModel, get_region_max, get_region_argmax


def calc_misfit(wf1, wf2, inds_only=None, exclude_inds=None):
    if inds_only is not None and exclude_inds is not None:
        raise RuntimeError("Error calculating misfit. ")

    inds_only = np.asarray(inds_only) if inds_only is not None else None
    exclude_inds = np.asarray(exclude_inds) if exclude_inds is not None else None

    if exclude_inds is not None and exclude_inds.size:
        mask = np.ones(wf1.size, dtype=bool)
        mask[exclude_inds] = False
        misfit = 100 * np.sqrt(np.mean((wf1[mask] - wf2[mask]) ** 2))
    elif inds_only is not None and inds_only.size:
        # according to Dinardo, 2020, eq. 3.52
        misfit = 100 * np.sqrt(np.mean((wf1[inds_only] - wf2[inds_only]) ** 2))
    else:
        # according to Dinardo, 2020, eq. 3.52
        misfit = 100 * np.sqrt(np.mean((wf1 - wf2) ** 2))

    return misfit


def is_two_step_retracking_active_samplus(entropy, pulse_peakiness, misfit):
    """Checks for the four conditions as defined in Dinardo2020 Eq. 3.36

    :param entropy: entropy of the waveform as defined in Dinard2020 Eq. 3.37
    :param pulse_peakiness: pulse peakiness as defined in Dinardo2020 Eq. 3.38
    :param misfit: misfit between modelled and received waveform as defined in Dinardo2020 Eq. 3.52
    :returns: True if one of the conditions are met
    """
    return np.logical_or.reduce(
        (
            np.multiply(entropy, pulse_peakiness) < 0.68,
            np.multiply(entropy, pulse_peakiness) > 0.78,
            (100 * pulse_peakiness) > 4.0,
            (entropy / misfit) < 8.0,
        )
    )


def longest_consecutive_run(data):
    return sorted(
        np.split(data, np.where(np.diff(data) != 1)[0] + 1), key=len, reverse=True
    )[0]


def oversample_wf(wf, int_oversampling_fac):
    """Linearly upsamples waveform. Samples inbetween are interpolated.
    The very last (int_oversampling_fac - 1) are replicated from the very last interpolated sample

    :param wf: waveform to be oversampled
    :param int_oversampling_fac: internal oversampling factor, must be even-numbered
    :return: the upsampled waveform with length len(wf) * int_oversampling_fac
    """
    wf_len = len(wf)
    x = np.arange(wf_len)
    interpolator = Akima1DInterpolator(x, wf)
    # interpolator = PchipInterpolator(x, wf)
    x_interp = np.arange(
        0, wf_len - 1 + 1 / int_oversampling_fac, 1 / int_oversampling_fac
    )
    wf_oversampled = np.full(wf_len * int_oversampling_fac, np.nan)
    wf_oversampled[: -(int_oversampling_fac - 1)] = interpolator(x_interp)
    # replace last four values by the last valid ones
    wf_oversampled[-(int_oversampling_fac):] = wf_oversampled[-(int_oversampling_fac)]

    return wf_oversampled


class SamosaRetracker:
    def __init__(
        self,
        retracker_basetype: RetrackerBaseType,
        retrack_sets: RetrackerSettings,
        fitting_sets: FittingSettings,
        sensor_sets: SensorSettings,
        wf_sets: WaveformSettings,
    ):
        self.retrack_sets = retrack_sets
        self.fitting_sets = fitting_sets
        self.sensor_sets = sensor_sets
        self.wf_sets = wf_sets

        self.model_sets = ModelSettings.get_default_sets(
            st=sensor_sets.sensor_type,
            retracker_basetype=retracker_basetype,
            n_effective_looks=self.retrack_sets.n_effective_looks,
            wf_sets=self.wf_sets,
        )

        if self.retrack_sets.fit_zero_doppler:
            self.model_sets.fit_zero_doppler = True

        # additional helper information
        self.B_r_Hz_oversampled = (
            self.sensor_sets.B_r_Hz
            * self.wf_sets.zp_oversampling_factor
            * self.wf_sets.internal_oversampling_factor
        )

    def _check_alt_flags(self):
        AO_Flag = True
        AM_Flag = True
        WQ_Flag = True

        return AO_Flag and AM_Flag and WQ_Flag

    @property
    def retracker_basetype(self):
        return self.retrack_sets.retracker_basetype

    @retracker_basetype.setter
    def retracker_basetype(self, rt_type):
        self.retrack_sets.retracker_basetype = rt_type
        self.model_sets.retracker_basetype = rt_type

    def get_wf_max(wf, retrack_sets, fg_epoch=0):
        if retrack_sets.normalise_wf_by_fg_region:
            wf_max = get_region_max(
                wf, fg_epoch, retrack_sets.normalise_wf_by_fg_region
            )
        else:
            # Wf_in_Norm_part = wf[retrack_sets.Wf_Norm_First - 1:retrack_sets.Wf_Norm_Last]
            # wf_max = np.max(pd.Series(Wf_in_Norm_part).rolling(retrack_sets.Wf_Norm_Aw).mean().dropna().tolist())
            wf_max = np.max(wf)

        return wf_max

    def apply_fit_mask(*, wf, wf_sets, retrack_sets):
        os_factor = wf_sets.zp_oversampling_factor
        first_bin = (
            retrack_sets.Wf_Fit_First_Bin * os_factor
            if retrack_sets.Wf_Fit_First_Bin != -1
            else None
        )
        last_bin = (
            retrack_sets.Wf_Fit_Last_Bin * os_factor
            if retrack_sets.Wf_Fit_Last_Bin != -1
            else None
        )

        if first_bin:
            wf[:first_bin] = 0
        if last_bin:
            wf[last_bin + 1 :] = 0

        return wf

    def init_processing(self, l1b_data_single, *, disable_oversampling=False):
        Wf_in = l1b_data_single["wf"].squeeze()

        # perform an internal oversampling of the waveform to ease fitting
        # procedure
        int_oversampling_fac = self.wf_sets.internal_oversampling_factor
        if int_oversampling_fac > 1 and not disable_oversampling:
            Wf_in = oversample_wf(Wf_in, self.wf_sets.internal_oversampling_factor)

        if self.retrack_sets.n_effective_looks != 0 and "stack" in l1b_data_single:
            n_total_looks = l1b_data_single["stack"].shape[1]
            first_look = n_total_looks // 2 - self.retrack_sets.n_effective_looks // 2
            inds_to_sum = np.arange(
                first_look, first_look + self.retrack_sets.n_effective_looks
            )

            Wf_in = np.nansum(l1b_data_single["stack"][:, inds_to_sum], axis=1)

            if int_oversampling_fac > 1 and not disable_oversampling:
                Wf_in = oversample_wf(Wf_in, self.wf_sets.internal_oversampling_factor)

        elif (
            self.retrack_sets.n_effective_looks != 0 and "stack" not in l1b_data_single
        ):
            raise RuntimeError(
                "No stack data exists although n_effective_looks param is configured!"
            )

        if np.isnan(Wf_in).any():
            raise RuntimeError("Wf_in contains NaN values.")
        if (Wf_in.size) != self.wf_sets.np:
            raise RuntimeError(f"Wf_in must be of length Np(={self.wf_sets.np})")

        # Additional sanity checks on l1b_data_single
        if len(l1b_data_single["beam_ang_stack_rad"]) and np.all(
            np.isnan(l1b_data_single["beam_ang_stack_rad"])
        ):
            raise RuntimeError("Invalid beam_ang_stack_rad variable. ")

        self.dynamic_fg_epoch = None
        if "dynamic_fg_epoch" in l1b_data_single:
            self.dynamic_fg_epoch = (
                l1b_data_single["dynamic_fg_epoch"] * int_oversampling_fac
            )

        if (
            self.retrack_sets.normalise_wf_by_fg_region
            and "dynamic_fg_epoch" not in l1b_data_single
        ):
            raise RuntimeError("Invalid config. ")

        if "entropy" in l1b_data_single:
            self.entropy = l1b_data_single["entropy"]

        if "pulse_peakiness" in l1b_data_single:
            self.pulse_peakiness = l1b_data_single["pulse_peakiness"]

        # I1: Initialize output variables to default values
        # I2: Check altimeter flags to decide if conditions are met to proceed
        # with fitting
        if not self._check_alt_flags():
            raise RuntimeError("Wf_Fit_Proceed_Flag = False")

        # I3: Estimate the Thermal Noise
        os_factor = self.wf_sets.zp_oversampling_factor
        Wf_in_TN_part = Wf_in[
            (self.retrack_sets.Wf_TN_First * os_factor) : (
                self.retrack_sets.Wf_TN_Last * os_factor
            )
            + 2
        ]

        # in case of SAMPLUS-L1B waveforms, add 4.84 dB to thermal noise,
        # Dinardo2020 3.2.3
        if (
            (
                self.retracker_basetype == RetrackerBaseType.SAMPLUS
                or self.retracker_basetype == RetrackerBaseType.SAMPLUSPLUS
            )
            and "thermal_noise" in l1b_data_single
            and not np.isnan(l1b_data_single["thermal_noise"])
        ):
            self.TN = l1b_data_single["thermal_noise"]
        else:
            self.TN = np.mean(Wf_in_TN_part)

        # I4: Compute the waveform  maximum value to be used for normalisation
        if self.retrack_sets.Wf_Norm_Flag:
            if self.dynamic_fg_epoch:
                self.Wf_max = SamosaRetracker.get_wf_max(
                    Wf_in, self.retrack_sets, fg_epoch=self.dynamic_fg_epoch
                )
            else:
                self.Wf_max = SamosaRetracker.get_wf_max(Wf_in, self.retrack_sets)
            # The value of Wf_max needs to be checked against TN to determine
            # if normalisation can proceed.
            if self.Wf_max <= self.TN * self.retrack_sets.Thr:
                self.Wf_norm = Wf_in
                raise RuntimeError("Wf_max <= TN * Thr")

            # I5: Apply normalisation to the waveform
            self.Wf_norm = Wf_in / self.Wf_max
            self.TN = self.TN / self.Wf_max
        else:
            self.Wf_norm = Wf_in
            self.Wf_max = 1.0

        # adjust Wf_Fit_First_Bin/Wf_Fit_Last_Bin according to RAW or RMC mode
        if (
            "s6" in self.sensor_sets.sensor_type.value
            and self.retrack_sets.auto_detect_s6_rmc
        ):
            is_rmc_active = np.mean(self.Wf_norm[-self.wf_sets.np // 3 :]) < 1e-3

            if is_rmc_active:
                self.retrack_sets.Wf_Fit_First_Bin = 11
                self.retrack_sets.Wf_Fit_Last_Bin = 132
            else:
                self.retrack_sets.Wf_Fit_First_Bin = 1
                self.retrack_sets.Wf_Fit_Last_Bin = self.wf_sets.np

        if np.allclose(self.Wf_norm, 0.0):
            raise RuntimeError("Wf_in are all-zero after normalisation. ")

        # I6: Check user control flag Wf_Fit_Activate_Flag to see if the user wants to proceed with waveform fitting
        # if self.sets.Wf_Fit_Activate_Flag:

    def _get_least_square_weights(self):
        wf = self.Wf_norm
        ls_weights = np.ones(wf.shape[0])
        le_inds = None

        # leading edge weighting factor
        try:
            le_inds = detect_leading_edge(
                wf,
                fg_epoch=self.dynamic_fg_epoch,
                normalise_wf_by_fg_region=self.retrack_sets.normalise_wf_by_fg_region,
            )
            ls_weights[le_inds] *= self.retrack_sets.leading_edge_weight_factor
        except Exception as e:
            logging.debug(f"Error detecting leading edge: {e}")

        return ls_weights, le_inds

    def _fun_model(p, *args):
        kwargs = args[0]
        _self = kwargs["self"]
        wf = _self.Wf_norm
        retrack_sets = _self.retrack_sets
        model_params = kwargs["model_params"]
        fitting_params = kwargs["fitting_params"]
        ls_weights = kwargs["ls_weights"]

        epoch_init = p[0]

        logging.debug(
            f"Generating SAMOSA waveform with params: P_u={p[2]}, Hs={p[1]}, t0_ns={epoch_init}"
        )

        # return diff wf_model - wf if requested
        do_return_diff = (
            True if "return_diff" in kwargs and kwargs["return_diff"] else False
        )

        sm = SamosaModel(
            model_sets=_self.model_sets,
            sensor_sets=_self.sensor_sets,
            wf_sets=_self.wf_sets,
            # reset the internal_oversampling_factor and generates a
            # non-oversampled model
        )

        # FIT5. Determine the theoretical model waveform to be used for the
        # fitting of the waveform
        try:
            if fitting_params.fit_nu_instead_of_swh:
                if retrack_sets.Disable_ML_Flag:
                    wf_model = sm.get_waveform_singlelook(
                        Pu=p[2],
                        Hs=fitting_params.swh_first_step,
                        t0_ns=epoch_init,
                        fa_Hz=0,
                        nu=p[1],
                        model_params=model_params,
                    )
                else:
                    wf_model = sm.get_waveform_multilook(
                        Pu=p[2],
                        Hs=fitting_params.swh_first_step,
                        t0_ns=epoch_init,
                        nu=p[1],
                        model_params=model_params,
                    )
            else:
                if retrack_sets.Disable_ML_Flag:
                    wf_model = sm.get_waveform_singlelook(
                        Pu=p[2],
                        Hs=p[1],
                        t0_ns=epoch_init,
                        fa_Hz=0,
                        model_params=model_params,
                    )
                else:
                    wf_model = sm.get_waveform_multilook(
                        Pu=p[2], Hs=p[1], t0_ns=epoch_init, model_params=model_params
                    )
        except RuntimeWarning as e:
            logging.warning(
                f'Warning in generating multilooked waveform: {" ".join(e.args)}'
            )
            wf_model = np.zeros(_self.wf_sets.np)

        if np.allclose(wf_model, 0.0):
            logging.warning(
                f"Calculated multilooked waveform is zero. Params: P_u={p[2]}, Hs={p[1]}, t0_ns={epoch_init}"
            )
            _self.res_fit_info["wf_opt"] = wf_model
            return wf_model

        wf_model += _self.TN
        diff = wf_model - wf

        diff *= ls_weights

        if (
            retrack_sets.interference_masking
            and not fitting_params.disable_interference_masking
        ):
            sm.wf_sets = _self.wf_sets  # use oversampled model waveform in this case

            interference_mask, interference_inds = sm.get_waveform_interference_mask(
                wf=wf,
                t0_ns=epoch_init,
                model_params=model_params,
                fg_epoch_gates=_self.dynamic_fg_epoch,
                n_normalise_grow=retrack_sets.normalise_wf_by_fg_region,
                n_mask_grow=retrack_sets.interference_masking_grow,
                swh_max=retrack_sets.interference_masking_swh_max,
                mask_before_le=retrack_sets.interference_masking_mask_before_le,
            )

            # remove overweighted indices from ls_weights
            le_inds = (ls_weights > 1.0).nonzero()[0]
            interference_inds = [ei for ei in interference_inds if ei not in le_inds]

            diff[interference_inds] = 0.0
            _self.res_fit_info["interference_inds"] = interference_inds
            _self.res_fit_info["interference_mask"] = interference_mask

        _self.res_fit_info["wf_opt"] = wf_model
        _self.res_fit_info["wf_model_max"] = sm.model_max

        # subwaveform approach
        if _self.retrack_sets.subwaveform_mode:
            max_le_gate = get_region_argmax(
                wf,
                region_center=_self.dynamic_fg_epoch,
                n_before_after=_self.retrack_sets.normalise_wf_by_fg_region,
            )
            diff[max_le_gate + _self.retrack_sets.subwaveform_n_gates_after_le :] = 0.0

            _self.res_fit_info["max_le_gate"] = max_le_gate

        res = diff if do_return_diff else wf_model
        res = SamosaRetracker.apply_fit_mask(
            wf=res, wf_sets=_self.wf_sets, retrack_sets=_self.retrack_sets
        )

        return res

    def _run_fitting_process(
        self, model_params: ModelParameter, fitting_params: FittingParameters = None
    ):
        fitting_params = (
            fitting_params if fitting_params is not None else FittingParameters()
        )

        fg_gate = (
            self.dynamic_fg_epoch if self.dynamic_fg_epoch else np.argmax(self.Wf_norm)
        )

        # FIT4. Determine the variables to be fitted, the number of variables to be fitted and their initial values
        # Vars to be fitted: [t0, Hs, Pu]

        # epoch/t0
        minmax_epoch = (
            fitting_params.minmax_epoch
            if fitting_params.minmax_epoch
            else self.fitting_sets.Fit_Var_1_MinMax_epoch
        )
        init_epoch = (
            fitting_params.init_epoch[0]
            if fitting_params.init_epoch
            else self.t[fg_gate] / 1e-9
        )

        # Hs
        if fitting_params.fit_nu_instead_of_swh:
            minmax_hs_or_nu = self.fitting_sets.Fit_Var_5_MinMax_nu
            init_hs_or_nu = self.fitting_sets.Fit_Var_5_Init_nu
        else:
            minmax_hs_or_nu = (
                fitting_params.minmax_swh
                if fitting_params.minmax_swh
                else self.fitting_sets.Fit_Var_2_MinMax_Hs
            )
            init_hs_or_nu = (
                fitting_params.init_swh
                if fitting_params.init_swh
                else self.fitting_sets.Fit_Var_2_Init_Hs
            )

        # Pu
        minmax_pu = (
            fitting_params.minmax_pu
            if fitting_params.minmax_pu
            else self.fitting_sets.Fit_Var_3_MinMax_Pu
        )
        init_pu = (
            fitting_params.init_pu
            if fitting_params.init_pu
            else self.fitting_sets.Fit_Var_3_Init_Pu
        )

        Fit_Var_Init = np.array([init_epoch, init_hs_or_nu, init_pu])

        # FIT5. Determine the theoretical model waveform to be used for the fitting of the waveform
        # FIT6. Computation of the new estimates by non-linear least-square fit
        ls_weights, le_inds = self._get_least_square_weights()

        _fun_model_arg = (
            {
                "self": self,
                "model_params": model_params,
                "fitting_params": fitting_params,
                "ls_weights": ls_weights,
                "subwaveform_stopgate": 0,
            },
        )

        self.res_fit_info = {
            "interference_inds": np.array([]),
            "interference_mask": np.ones(self.Wf_norm.shape[0]),
        }

        logging.debug(
            f"Started fitting. Init values: swh={init_hs_or_nu:.2f}m,"
            f"epoch={init_epoch:.2f}ns, Pu={init_pu:.2f}W"
        )

        ns_per_gate = self.dtau * 1e9
        epoch_min, epoch_max = minmax_epoch[0], minmax_epoch[1]

        if self.dynamic_fg_epoch and self.fitting_sets.limit_epoch_around_fg_n:
            n_gates_before_after = self.fitting_sets.limit_epoch_around_fg_n
            epoch_min = init_epoch - n_gates_before_after * ns_per_gate
            epoch_max = init_epoch + n_gates_before_after * ns_per_gate

        Fit_Var_Bounds = (
            [epoch_min, minmax_hs_or_nu[0], minmax_pu[0]],
            [epoch_max, minmax_hs_or_nu[1], minmax_pu[1]],
        )

        max_nfev = self.fitting_sets.Fit_MaxIter

        res_ls = least_squares(
            fun=SamosaRetracker._fun_model,
            x0=Fit_Var_Init,
            bounds=Fit_Var_Bounds,
            method="trf",
            args=({**_fun_model_arg[0], **{"return_diff": True}},),
            ftol=self.fitting_sets.trf_threshold,
            gtol=self.fitting_sets.trf_threshold,
            xtol=self.fitting_sets.trf_threshold,
            diff_step=self.fitting_sets.trf_stepsize,
            max_nfev=max_nfev,
            # verbose=2,
            # **loss_kw,
        )

        if not res_ls.success:
            RuntimeError(f"Failed TRF fitting: {res_ls.message}. ")

        if fitting_params.fit_nu_instead_of_swh:
            res_fit = {
                "swh": fitting_params.swh_first_step,
                "epoch_ns": res_ls.x[0],
                "Pu": res_ls.x[2],
                "nu": res_ls.x[1],
                "n_iter": res_ls.nfev,
                **self.res_fit_info,
            }
        else:
            res_fit = {
                "swh": res_ls.x[1],
                "epoch_ns": res_ls.x[0],
                "Pu": res_ls.x[2],
                "nu": 0,
                "n_iter": res_ls.nfev,
                **self.res_fit_info,
            }

        self.res_fit_info["wf_opt"] = SamosaRetracker.apply_fit_mask(
            wf=self.res_fit_info["wf_opt"],
            wf_sets=self.wf_sets,
            retrack_sets=self.retrack_sets,
        )

        # take fitted waveform that was generated at the very last step
        wf_opt = (
            self.res_fit_info["wf_opt"]
            if "wf_opt" in self.res_fit_info
            else np.zeros(len(self.Wf_norm))
        )

        # FIT7: Assign fitted variables to the output values
        # (FIT8. Re-convert Epoch in seconds)
        # FIT9. De-normalise the retrieved Power and Noise Values
        factor_denorm = self.Wf_max / self.res_fit_info["wf_model_max"]

        misfit = calc_misfit(wf_opt, self.Wf_norm)
        if self.retrack_sets.subwaveform_mode:
            misfit_selective = calc_misfit(
                wf_opt[
                    : res_fit["max_le_gate"]
                    + self.retrack_sets.subwaveform_n_gates_after_le
                ],
                self.Wf_norm[
                    : res_fit["max_le_gate"]
                    + self.retrack_sets.subwaveform_n_gates_after_le
                ],
            )
        else:
            misfit_selective = calc_misfit(
                wf_opt,
                self.Wf_norm,
                exclude_inds=self.res_fit_info["interference_inds"],
            )

        return {
            **res_fit,
            **{
                "wf": self.Wf_norm,
                "wf_opt": wf_opt,
                "Pu_denorm": res_fit["Pu"] * factor_denorm,
                "factor_denorm": factor_denorm,
                "misfit": misfit,
                "misfit_le": calc_misfit(wf_opt, self.Wf_norm, inds_only=le_inds),
                "misfit_selective": misfit_selective,
                "le_inds": le_inds,
                "l2norm": np.linalg.norm(self.Wf_norm - wf_opt),
            },
        }

    def fit_wf(self, l1b_data_single, model_params: ModelParameter):
        self.init_processing(l1b_data_single)

        # FIT1. Initialize output variables to default values
        # FIT2. Determine if waveform fitting should proceed
        # FIT3. Calculate the waveform time abscissa
        self.dtau = 1 / self.B_r_Hz_oversampled
        epoch_ref_gate = (
            model_params.epoch_ref_gate * self.wf_sets.internal_oversampling_factor
        )
        self.t = (
            np.arange(-epoch_ref_gate, -epoch_ref_gate + self.wf_sets.np) * self.dtau
        )

        # FIT4-9
        res_fit = self._run_fitting_process(model_params)

        if self.retrack_sets.interference_masking_second_retracking_step:
            n_iter_first = res_fit["n_iter"]

            logging.debug(
                f'Starting second iteration step with updated interference_masking_swh_max={res_fit["swh"]}. '
            )
            upper_swh_margin = 2.0
            self.retrack_sets.interference_masking_swh_max = np.min(
                [
                    res_fit["swh"] + upper_swh_margin,
                    self.fitting_sets.Fit_Var_2_MinMax_Hs[1],
                ]
            )
            res_fit = self._run_fitting_process(model_params)

            res_fit["n_iter"] = (n_iter_first, res_fit["n_iter"])

        # SAMOSA+ addition: second retracking step
        if self.retrack_sets.second_retracking_step_samplus:
            if (
                "entropy" in dir(self)
                and "pulse_peakiness" in dir(self)
                and is_two_step_retracking_active_samplus(
                    entropy=self.entropy,
                    pulse_peakiness=self.pulse_peakiness,
                    misfit=res_fit["misfit"],
                )
            ):
                n_iter_first = res_fit["n_iter"]

                fitting_params = FittingParameters(
                    fit_nu_instead_of_swh=True,
                    swh_first_step=res_fit["swh"],
                    disable_interference_masking=False,
                )
                logging.debug(
                    f'Starting second iteration step, taking SWH={res_fit["swh"]} from first step. '
                )
                res_fit = self._run_fitting_process(
                    model_params=model_params, fitting_params=fitting_params
                )

                res_fit["n_iter"] = (n_iter_first, res_fit["n_iter"])

        logging.debug(
            f'Finished fitting (n_iter={res_fit["n_iter"]}). Found params: swh={res_fit["swh"]:.2f}m, '
            f'epoch={res_fit["epoch_ns"]:.2f}ns, Pu={res_fit["Pu"]:.2f}W, nu={res_fit["nu"]:.2e}. '
        )

        return res_fit



from bisect import bisect_left
from itertools import chain

import numpy as np

from pysamosa.common_types import (
    ModelSettings,
    ModelParameter,
    RetrackerBaseType,
    WaveformSettings,
    SensorSettings,
)
from pysamosa.conf_params import CONST_A, CONST_B, CONST_C, CONST_F
from pysamosa.model_helpers import load_samosa_luts, get_f_from_lut
from pysamosa.rip import RipAnalyser


def get_region_max(wf, region_center, n_before_after):
    return np.max(wf[region_center - n_before_after : region_center + n_before_after])


def get_region_argmax(wf, region_center, n_before_after):
    return np.argmax(
        wf[region_center - n_before_after : region_center + n_before_after]
    ) + (region_center - n_before_after)


def get_closest_idxs(array, values):
    array = np.array(array)
    idxs = np.searchsorted(array, values, side="left")  # get insert positions

    # find indexes where previous index is closer
    prev_idx_is_less = (idxs == len(array)) | (
        np.fabs(values - array[np.maximum(idxs - 1, 0)])
        < np.fabs(values - array[np.minimum(idxs, len(array) - 1)])
    )
    idxs[prev_idx_is_less] -= 1

    return idxs


class SamosaModel:
    def __init__(
        self,
        model_sets: ModelSettings,
        sensor_sets: SensorSettings,
        wf_sets: WaveformSettings,
    ):
        self.sensor_sets = sensor_sets
        self.model_sets = model_sets
        self.wf_sets = wf_sets

        self.B_r_Hz_oversampled = (
            self.sensor_sets.B_r_Hz
            * self.wf_sets.zp_oversampling_factor
            * self.wf_sets.internal_oversampling_factor
        )
        self.luts_samosa = load_samosa_luts()[self.sensor_sets.sensor_type.value]

    def _check_param(self, Pu: float, Hs: float, t0_ns: float):
        if not self.model_sets.Disable_LUT_Flag and not (
            self.luts_samosa["lut_alpha_p"][0][0]
            <= Hs
            <= self.luts_samosa["lut_alpha_p"][-1][0]
        ):
            raise RuntimeError(
                f"Hs must be within range [{self.luts_samosa['lut_alpha_p'][0][0]}, {self.luts_samosa['lut_alpha_p'][-1][0]}]"
            )

    def get_waveform_singlelook(
        self,
        Pu: float,
        Hs: float,
        t0_ns: float,
        fa_Hz: float,
        nu: float = 0.0,
        *,
        model_params,
        custom_gamma0=None,
    ):
        """
        :param Pu: waveform power amplitude
        :param Hs: significant wave height
        :param t0_ns: epoch relative to central gate in ns
        :param fa_Hz: Doppler frequency at which to calculate the single-look theoretical delay waveform
        :param model_params: ModelParameter() object with the model configuration
        :return:
        """

        self._check_param(Pu=Pu, Hs=Hs, t0_ns=t0_ns)

        # SL1. Initialize output variables to default values
        Np = int(self.wf_sets.np) - (self.wf_sets.internal_oversampling_factor - 1)
        h = model_params.alt_m

        # SL2. Compute local Earth radius, Re, and spherical surface parameter, alpha
        # Compute the local radii of curvature of the Earth's surface (Maulik Jain,
        # Improved sea level determination in the Arctic regions through development
        # of tolerant altimetry retracking, Eq. 5.2, pp. 47)
        # lat_geocentric = np.arctan((1 - CONST_F)**2 * np.tan(model_params.lat_rad))
        # Re = np.sqrt(CONST_A**2 * np.cos(lat_geocentric)**2 + CONST_B**2 *
        # np.sin(lat_geocentric)**2)  #more accurate (lat-dependent) calculus
        # of Earth' radius
        Re = np.sqrt(
            CONST_A**2 * np.cos(model_params.lat_rad) ** 2
            + CONST_B**2 * np.sin(model_params.lat_rad) ** 2
        )
        alpha = 1 + h / Re

        # SL3. Calculate model_paramsmodel_params.lat_radlat_rad new parameters
        prf_hz = (
            self.sensor_sets.prf_Hz
            if self.sensor_sets.prf_Hz is not None
            else 1 / model_params.pri_hz
        )
        # lambda0 = c / self.sensor_sets['f_c_Hz']
        dtau = 1 / self.B_r_Hz_oversampled
        dfa = prf_hz / self.sensor_sets.n_b
        xp = h * np.tan(model_params.ksix_rad)
        yp = -h * np.tan(model_params.ksiy_rad)
        alphax = (
            self.sensor_sets.sh_x
            * 8
            * np.log(2)
            / (h**2 * self.sensor_sets.theta_x_rad**2)
        )
        alphay = (
            self.sensor_sets.sh_y
            * 8
            * np.log(2)
            / (h**2 * self.sensor_sets.theta_y_rad**2)
        )
        Tb = self.sensor_sets.n_b / prf_hz
        Lx = (CONST_C * h) / (
            2 * model_params.Vs_m_per_s * self.sensor_sets.f_c_Hz * Tb
        )
        Ly = np.sqrt((CONST_C * h) / (alpha * self.sensor_sets.B_r_Hz))
        Lz = CONST_C / (2 * self.sensor_sets.B_r_Hz)
        Lg = alpha / (2 * h * alphay)
        sigmaz = Hs / 4

        # SL4. Define indices in delay (K) and doppler (L) space
        epoch_ref_gate = (
            model_params.epoch_ref_gate * self.wf_sets.internal_oversampling_factor
        )

        t_start_val = -epoch_ref_gate
        t = (np.arange(t_start_val, t_start_val + Np)) * dtau
        tau = t - float(t0_ns) * 1e-9
        K = tau * self.sensor_sets.B_r_Hz
        L = int(np.round(fa_Hz / dfa))

        # SL5. Define a value for alph_P
        if (
            self.model_sets.retracker_basetype == RetrackerBaseType.SAMPLUS
            and 0.0 <= model_params.dist2coast < 10.0
        ):
            alpha_p = 0.55
        elif self.model_sets.Disable_LUT_Flag:
            alpha_p = self.model_sets.alpha_p_mean
        else:
            alpha_p = self.luts_samosa["lut_alpha_p"][
                bisect_left(self.luts_samosa["lut_alpha_p"][:, 0], Hs), 1
            ]

        # SL5. Calculate GL and GLK
        track_sign = -1 if model_params.ascending else 1
        orbit_slope = track_sign * (
            (CONST_A**2 - CONST_B**2) / (2 * Re**2)
        ) * np.sin(2 * model_params.lat_rad) - (
            -model_params.h_rate_m_per_s / model_params.Vs_m_per_s
        )
        Ls = orbit_slope * h / (alpha * Lx)
        # for S6: Enable_Slope_Effect_Flag is inherently turned on because a
        # different GL formula is used
        if self.model_sets.Enable_Slope_Effect_Flag:
            gamma = 2 * (L - Ls) * Lx**2 / Ly**2
        else:
            gamma = 2 * L * Lx**2 / Ly**2

        sign = Hs / np.abs(Hs) if Hs != 0 else 1

        # sigma_p = alpha_p / self.B_r_Hz_oversampled
        # theta_l = Lx / h * (L - Ls)
        # GL = 1 / (self.B_r_Hz_oversampled * np.sqrt(sigma_p**2 * (1 + (Lx /
        # Lz)**2 * (alpha * theta_l)**2) + (2 * sigmaz / CONST_C)**2))
        # #alternative formula from Dinardo et al. 2018
        GL = 1 / np.sqrt(
            alpha_p**2 + alpha_p**2 * gamma**2 + sign * (Hs / (4 * Lz)) ** 2
        )

        GLK = GL * K

        # SL6. Calculate GAMMA0 and Const
        XL = L * Lx
        YK = np.zeros(Np)
        YK[K > 0] = Ly * np.sqrt(K[K > 0])

        if (
            self.model_sets.retracker_basetype == RetrackerBaseType.SAMPLUSPLUS
            and custom_gamma0 is not None
        ):
            gamma0 = custom_gamma0
        else:
            gamma0 = np.exp(
                -alphay * yp**2
                - alphax * (XL - xp) ** 2
                - XL**2 * nu / h**2
                - (alphay + nu / h**2) * YK**2
            ) * np.cosh(2 * alphay * yp * YK)
        # here it is intended to use the constant alpha_p term
        Const = self.model_sets.alpha_p_mean**2 * np.sqrt(2 * np.pi)

        # SL6. Calculate Tk
        Tk = np.zeros(Np)
        Tk[K > 0] = (1 + nu / (alphay * h**2)) - (
            yp / (Ly * np.sqrt(K[K > 0]))
        ) * np.tanh(2 * alphay * yp * Ly * np.sqrt(K[K > 0]))
        Tk[K <= 0] = (1 + nu / (alphay * h**2)) - 2 * alphay * yp**2

        # SL7. Calculate the final expression of Pr_SL
        f_0_l_k = get_f_from_lut(GLK, self.luts_samosa["lut_F0"], 0)
        f_1_l_k = get_f_from_lut(GLK, self.luts_samosa["lut_F1"], 1)

        if self.model_sets.Disable_SAM_F1_Flag:
            Pr_SL = Const * np.sqrt(GL) * gamma0 * f_0_l_k
        else:
            Pr_SL = (
                Const
                * np.sqrt(GL)
                * gamma0
                * (f_0_l_k + (sigmaz / Lg) * (sigmaz / Lz) * GL * Tk * f_1_l_k)
            )

        return Pr_SL

    def get_waveform_multilook(self, Pu, Hs, t0_ns, nu=0, *, model_params=None):
        model_params = model_params if model_params else ModelParameter()

        self._check_param(Pu=Pu, Hs=Hs, t0_ns=t0_ns)

        prf_hz = (
            self.sensor_sets.prf_Hz
            if self.sensor_sets.prf_Hz is not None
            else 1 / model_params.pri_hz
        )
        # ML1. Initialize output variables to default values

        # ML2. Calculate idealised Doppler beam angles if not available from
        # the SAR L1B input file
        h = model_params.alt_m
        Vs = model_params.Vs_m_per_s
        Tb = self.sensor_sets.n_b / prf_hz
        Np = int(self.wf_sets.np) - (self.wf_sets.internal_oversampling_factor - 1)

        # Compute the local radii of curvature of the Earth's surface (Maulik Jain,
        # Improved sea level determination in the Arctic regions through development
        # of tolerant altimetry retracking, Eq. 5.2, pp. 47)
        np.arctan((1 - CONST_F) ** 2 * np.tan(model_params.lat_rad))
        # Re = np.sqrt(CONST_A**2 * np.cos(lat_geocentric)**2 + CONST_B**2 * np.sin(lat_geocentric)**2)
        Re = np.sqrt(
            CONST_A**2 * np.cos(model_params.lat_rad) ** 2
            + CONST_B**2 * np.sin(model_params.lat_rad) ** 2
        )
        alpha = 1 + h / Re

        if self.model_sets.Ideal_Beam_Ang_Stack_flag or not len(
            model_params.beam_ang_stack_rad
        ):
            dTheta = (Vs * self.sensor_sets.bri) / (h * alpha)
            Theta1 = np.pi / 2 + dTheta * self.model_sets.Ideal_First_Look_Index_ML
            Theta2 = np.pi / 2 + dTheta * self.model_sets.Ideal_Last_Look_Index_ML
            self.Beam_Ang_Stack_TS = np.arange(Theta1, Theta2, dTheta)
        else:
            beam_ang_stack_full = model_params.beam_ang_stack_rad[
                ~np.isnan(model_params.beam_ang_stack_rad)
            ]

            # reduce beam_ang_stack_full
            n_looks_to_sum = self.model_sets.n_effective_looks
            # n_looks_to_sum = 0
            if n_looks_to_sum == 0:
                self.Beam_Ang_Stack_TS = model_params.beam_ang_stack_rad[
                    ~np.isnan(model_params.beam_ang_stack_rad)
                ]
            else:
                self.Beam_Ang_Stack_TS = beam_ang_stack_full[
                    beam_ang_stack_full.shape[0] // 2
                    - n_looks_to_sum // 2 : beam_ang_stack_full.shape[0] // 2
                    + n_looks_to_sum // 2
                ]

        # ML3. Calculate the vector of Doppler frequency of the Doppler beams
        # used for multi-looking
        lambda0 = CONST_C / self.sensor_sets.f_c_Hz
        Dopp_Freq_Stack_TS = (2 * Vs / lambda0) * np.cos(self.Beam_Ang_Stack_TS)

        # ML4. Sub-setting the Stack
        dfa = prf_hz / self.sensor_sets.n_b
        Beam_Indices_float = Dopp_Freq_Stack_TS / dfa
        factor = 1
        Beam_Index = (
            np.unique((np.round(Beam_Indices_float / factor)).astype(int)) * factor
        )
        Dopp_Freq_Stack_TS = Beam_Index * dfa

        # ML5. Build the Stack of Doppler beams prior to incoherent integration
        Neff = Dopp_Freq_Stack_TS.size
        self.Pr_Stack = np.zeros([Neff, Np])

        # Calculate GAMMA0
        custom_gamma0 = None
        if self.model_sets.retracker_basetype == RetrackerBaseType.SAMPLUSPLUS:
            ripa = RipAnalyser(
                model_params.rip,
                sensor_sets=self.sensor_sets,
                model_params=model_params,
            )
            self.rip_params = ripa.rip_params

            custom_gamma0 = ripa.get_gamma0(n_looks_eff=Neff, n_gates=Np)

        for j, fa in enumerate(Dopp_Freq_Stack_TS):
            if self.model_sets.fit_zero_doppler:
                fa = 0.0

            if custom_gamma0 is not None:
                self.Pr_Stack[j] = self.get_waveform_singlelook(
                    Pu=Pu,
                    Hs=Hs,
                    t0_ns=t0_ns,
                    fa_Hz=fa,
                    nu=nu,
                    model_params=model_params,
                    custom_gamma0=custom_gamma0[:, j],
                )
            else:
                self.Pr_Stack[j] = self.get_waveform_singlelook(
                    Pu=Pu,
                    Hs=Hs,
                    t0_ns=t0_ns,
                    fa_Hz=fa,
                    nu=nu,
                    model_params=model_params,
                )

        # ML6. Mask out from the Stack of Doppler beams the power bins located
        # beyond the radar window length
        Lx = (lambda0 * h) / (2 * Vs * Tb)
        Beam_Range = h * (np.sqrt(1 + alpha * (Lx * Beam_Index / h) ** 2) - 1)
        dr = np.arange(0, Np)[::-1]
        Window_Range = CONST_C / (2 * self.B_r_Hz_oversampled) * dr

        if model_params.stack_mask_start_stop is not None:
            inds_stack_mask_mapped = get_closest_idxs(
                np.sort(Beam_Indices_float), np.sort(Beam_Index)
            )
            stack_mask_start_stop_mapped = model_params.stack_mask_start_stop[
                inds_stack_mask_mapped
            ].astype(int)

            for i, mask_val in np.ndenumerate(stack_mask_start_stop_mapped):
                self.Pr_Stack[i, mask_val:] = np.nan

        else:
            for j in np.arange(0, Neff):
                for i in np.arange(0, Np):
                    if Beam_Range[j] > Window_Range[i]:
                        self.Pr_Stack[j, i] = np.nan

        # ML7. Apply Weighting Function to the Stack prior to incoherent
        # summation
        Stack_Weights = np.ones(self.Pr_Stack.shape[0])
        self.Pr_Stack *= Stack_Weights[:, np.newaxis]

        # ML8. Calculate multi-looked waveform by incoherent summation across
        # the Stack
        self.Pr_ML = 1 / Neff * np.nansum(self.Pr_Stack, axis=0)
        # for S6: if avoid_zeros_in_the_multilooking is active then ignore the nan values
        # self.Pr_ML = np.nanmean(self.Pr_Stack, axis=0)
        # self.Pr_ML = np.nan_to_num(self.Pr_ML)

        self.model_max = np.max(self.Pr_ML)
        self.Pr_ML = Pu * (self.Pr_ML / self.model_max)

        # logging.debug(f'Generated multilooked waveform with params P_u={Pu:.4f}, Hs={Hs:.4f}, t0_ns={t0_ns:.4f}, nu={nu:.2e}')

        if self.wf_sets.internal_oversampling_factor > 1:
            self.Pr_ML = np.append(
                self.Pr_ML,
                np.repeat(
                    self.Pr_ML[-1], (self.wf_sets.internal_oversampling_factor - 1)
                ),
            )

        return self.Pr_ML

    def get_waveform_interference_mask(
        self,
        *,
        wf,
        t0_ns,
        model_params,
        fg_epoch_gates,
        n_normalise_grow,
        n_mask_grow,
        swh_max=None,
        mask_before_le=False,
    ):
        """Estimates the waveforms interference mask based on very-large-sea-state envelope.

        :param wf: the waveform to retrack
        :param t0_ns: the retracked epoch t0 in ns (referring to the epoch_ref_gate)
        :param model_params: the ModelParameters object for this measurement
        :param fg_epoch_gates: the estimated first-guess-epoch in gates
        :param n_normalise_grow: number of gates before and after a waveform's peak gate to normalise the waveform
        :param n_mask_grow: let the interference mask grow before and after the waveform gate that is larger than the envelope
        :param swh_max: maximum SWH boundary, from which the interference mask is generated
        :return: the envelope mask and the indices of the "bad" gates
        """
        if n_normalise_grow:
            wf = wf / get_region_max(
                wf, region_center=fg_epoch_gates, n_before_after=n_normalise_grow
            )
        else:
            wf = wf / np.max(wf)

        mask_max = self.get_waveform_singlelook(
            Pu=1.0, Hs=swh_max, t0_ns=t0_ns, nu=0.00, model_params=model_params, fa_Hz=0
        )
        mask_max = np.append(
            mask_max,
            np.repeat(mask_max[-1], self.wf_sets.internal_oversampling_factor - 1),
        )
        mask_max = mask_max / np.max(mask_max)
        interference_ref_waveform = mask_max

        mask_close_to_zero = np.isclose(interference_ref_waveform, 0.0, atol=1e-2)

        # set everything before max to 1.0
        mask_max_max = np.argmax(mask_max)
        interference_ref_waveform[:mask_max_max] = 1.0

        if mask_before_le:
            interference_ref_waveform[mask_close_to_zero] = 0.0

        # add safety margin to noise floor
        upper_noise_floor_bound = 0.05
        interference_ref_waveform += upper_noise_floor_bound

        if mask_before_le:
            interference_ref_waveform[mask_close_to_zero] = upper_noise_floor_bound

        # get indices of gates, for which the waveform is greater than the
        # modelled envelope
        inds_bad = (
            (wf.squeeze() > interference_ref_waveform.squeeze()).nonzero()[0].tolist()
        )

        # grow region of each bad ind by n_grow
        if len(inds_bad):
            if n_mask_grow:
                inds_bad = [
                    [i for i in range(j - n_mask_grow, j + n_mask_grow + 1)]
                    for j in inds_bad
                ]
                # flatten list and get unique set of bad inds
                inds_bad = sorted(list(set(list(chain.from_iterable(inds_bad)))))
            else:
                # flatten list and get unique set of bad inds
                inds_bad = list(set(inds_bad))

        # remove bad inds that are before fg_epoch_gates
        # inds_bad = [i for i in inds_bad if i >= fg_epoch_gates]

        # remove bad inds that are too close to the fg_epoch, meaning fg_epoch
        # +- n_mask_grow
        inds_bad = [i for i in inds_bad if (np.abs(fg_epoch_gates - i) > n_mask_grow)]
        inds_bad = [
            i for i in inds_bad if (np.abs(fg_epoch_gates - i) > n_normalise_grow)
        ]

        # remove invalid indices
        inds_bad = [i for i in inds_bad if (0 <= i < wf.shape[0])]

        return interference_ref_waveform, inds_bad



import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path
import string

from pysamosa.common_types import (
    RetrackerBaseType,
    WaveformSettings,
    SensorSettings,
    L1bSourceType,
    SensorType,
)
from pysamosa.model import SamosaModel, ModelParameter, ModelSettings
from pysamosa.data_access import (
    _read_dataset_vars_from_ds,
    get_subset_dataset,
    get_model_param_obj_from_l1b_data,
    data_vars_s6,
)
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset
from pysamosa.utils import gen_first_true
from pysamosa import retracker

default_width_in = 5.79  # in inches
default_ratio = 0.5 + np.sqrt(5) / 2  # 1.618, golden ratio
default_figsize_in = [default_width_in, default_width_in / default_ratio]

# set_pgf_mode()

okabeito_colorblind_colors = {
    "orange": "#E69F00",
    "skyblue": "#56B4E9",
    "bluishgreen": "#009E73",
    "yellow": "#F0E442",
    "blue": "#0072B2",
    "vermilion": "#D55E00",
    "reddishpurple": "#CC79A7",
    "black": "#000000",
}

THESIS_BASEDIR = Path("/home/schlembach/fastdata/repos/phd-project-flo/thesis/figures/")

fig_list = {}
fig_scale_factor = 2
figwidth_textwidth = fig_scale_factor * default_figsize_in[0]
figheight_textwidth = fig_scale_factor * default_figsize_in[1]
default_figsize = [figwidth_textwidth, figheight_textwidth]
default_dpi = 400
default_subplot_name_kw = {
    "xycoords": "axes fraction",
    "xy": (0.02, 0.96),
    "textcoords": "offset pixels",
    "xytext": (0, 0),
    "horizontalalignment": "left",
    "verticalalignment": "top",
    "fontweight": "bold",
    "color": "black",
    "bbox": dict(
        boxstyle="square,pad=.2",
        facecolor="white",
        edgecolor="none",
        alpha=0.7,
        zorder=10,
    ),
}
default_legend_kw = {"fontsize": 8, "labelspacing": 0.20, "borderaxespad": 0.2}


def gen_multilooked_vary_swh():
    fig, ax = plt.subplots(figsize=[5, 3], dpi=default_dpi)
    fig_list["samosa_model_vary_swh"] = fig

    cols_it = iter(list(okabeito_colorblind_colors.values()))

    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S6_F04)
    model_sets = ModelSettings(retracker_basetype=RetrackerBaseType.SAM)
    sensor_sets = SensorSettings.get_default_sets(st=SensorType.S6_F06)
    sm = SamosaModel(model_sets=model_sets, wf_sets=wf_sets, sensor_sets=sensor_sets)

    swh = [0.25, 0.75, 1.0, 2.0, 5, 7, 9]
    Pu = 1.0
    t_0 = -250
    nu = 0

    epoch_ref_gate = 256
    model_params = ModelParameter(epoch_ref_gate=epoch_ref_gate)
    model_params.pri_hz = 1 / 9.2e3
    sensor_sets.bri = 1 / 140
    for s in swh:
        wf = sm.get_waveform_multilook(
            Pu=Pu, Hs=s, t0_ns=t_0, nu=nu, model_params=model_params
        )
        # wf = wf + np.random.randn(len(wf)) * 0.01 * np.nanmax(wf)  # NOISE
        # ADDITION (not in the SAMOSA documentation )
        print(
            f"SWH={s} m, argmax={wf.argmax()}, retracking_point={np.abs(wf - 0.8422).argsort()[1]}"
        )
        ax.plot(wf, label=f"SWH={s:.2f} m", linewidth=1, color=next(cols_it))

    # plot epoch reference gate vertical line
    ax.axvline(
        epoch_ref_gate,
        linestyle="--",
        linewidth="1.0",
        label="Epoch reference gate",
        color=next(cols_it),
    )

    # plot arrow between retracking point and epoch reference gates
    level_retracking_point = 0.8422
    x_retracking_point = np.abs(wf - level_retracking_point).argsort()[0]
    ax.annotate(
        "",
        xy=(x_retracking_point, level_retracking_point),
        xytext=(epoch_ref_gate, level_retracking_point),
        arrowprops=dict(arrowstyle="<|-|>", shrinkA=0, shrinkB=0, color="black"),
    )
    ax.annotate(
        "epoch",
        xy=(
            x_retracking_point + (epoch_ref_gate - x_retracking_point) / 2,
            level_retracking_point + 0.02,
        ),
        horizontalalignment="center",
    )

    x_pu, y_pu = 25, 1.0
    ax.annotate(
        "",
        xy=(x_pu, 0.0),
        xytext=(x_pu, y_pu),
        arrowprops=dict(arrowstyle="<|-|>", shrinkA=0, shrinkB=0, color="black"),
    )
    ax.annotate("$P_u$", xy=(x_pu - 5, y_pu / 2), horizontalalignment="right")

    ax.set_xlabel("Range gate [\\#]")
    ax.set_ylabel("Normalised power")
    ax.legend(**default_legend_kw)
    ax.grid()


def plot_single_retrackings():
    fig, axs = plt.subplots(
        1,
        4,
        figsize=[figwidth_textwidth, figwidth_textwidth / 4],
        dpi=default_dpi,
        sharey="row",
    )
    axs = axs.ravel()
    ax_it = iter(axs)
    subplot_name_it = iter(string.ascii_lowercase)

    first_it = gen_first_true()

    fig_list["samosa_retracked_estimates"] = fig

    list_scenarios = [
        # open ocean
        (
            46720,
            Path(
                "/nfs/DGFI145/C/work_flo/coastal_ffsar/orig_data/f06/P4_1B_HR_____/S6A_P4_1B_HR______20210413T182810_20210413T192223_20220514T131950_3253_015_213_106_EUM__REP_NT_F06.SEN6/measurement.nc"
            ),
        ),
        # coastal: peak after leading edge
        (
            8586,
            Path(
                "/nfs/DGFI145/C/work_flo/coastal_ffsar/orig_data/f06/P4_1B_HR_____/S6A_P4_1B_HR______20210502T223038_20210502T232547_20220505T002438_3309_017_196_098_EUM__REP_NT_F06.SEN6/measurement.nc"
            ),
        ),
        # coastal: peak after leading edge
        (
            8581,
            Path(
                "/nfs/DGFI145/C/work_flo/coastal_ffsar/orig_data/f06/P4_1B_HR_____/S6A_P4_1B_HR______20210502T223038_20210502T232547_20220505T002438_3309_017_196_098_EUM__REP_NT_F06.SEN6/measurement.nc"
            ),
        ),
        # coastal: multipeak scenario
        (
            46624,
            Path(
                "/nfs/DGFI145/C/work_flo/coastal_ffsar/orig_data/f06/P4_1B_HR_____/S6A_P4_1B_HR______20210413T182810_20210413T192223_20220514T131950_3253_015_213_106_EUM__REP_NT_F06.SEN6/measurement.nc"
            ),
        ),
    ]
    datavars_l1b = data_vars_s6["l1b"]
    # datavars_l2 = data_vars_eumetsat_s6['l2']

    l1b_src_type, retracker_basetype, settings_preset = (
        L1bSourceType.EUM_S6_F06,
        RetrackerBaseType.SAM,
        SettingsPreset.NONE,
    )

    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=retracker_basetype,
        settings_preset=settings_preset,
        l1b_src_type=l1b_src_type,
    )

    for ind, file_l1b in list_scenarios:
        ax = next(ax_it)

        grp = "data_20/ku" if "s6" in str(file_l1b).lower() else None
        l1b_data = _read_dataset_vars_from_ds(
            nc_filename=file_l1b,
            data_var_names=datavars_l1b,
            n_offset=ind,
            n_inds=2,
            group=grp,
        )
        l1b = get_subset_dataset(l1b_data, ind_offset=0)
        # l2_data = _read_dataset_vars_from_ds(nc_filename=file_l2, data_var_names=datavars_l2, n_offset=ind, n_inds=2, group=grp)
        # l2 = get_subset_dataset(l2_data, ind_offset=0)

        model_params = get_model_param_obj_from_l1b_data(l1b_data, ind=0)

        sr = retracker.SamosaRetracker(
            retracker_basetype=retracker_basetype,
            retrack_sets=retrack_sets,
            fitting_sets=fitting_sets,
            sensor_sets=sensor_sets,
            wf_sets=wf_sets,
        )

        # start fitting
        res_fit = sr.fit_wf(l1b_data_single=l1b, model_params=model_params)

        ax.plot(
            l1b["wf"] / np.max(l1b["wf"]),
            color=okabeito_colorblind_colors["blue"],
            label="measured waveform",
        )
        ax.plot(
            res_fit["wf_opt"],
            color=okabeito_colorblind_colors["orange"],
            label=f"fitted SAMOSA2 model,\n"
            f'$P_u$={res_fit["Pu"]:.2f}, SWH={res_fit["swh"]:.2f} m,\n'
            f'epoch={res_fit["epoch_ns"]:.2f} ns, misfit={res_fit["misfit"]:.2f}',
        )

        ax.annotate(f"({next(subplot_name_it)})", **default_subplot_name_kw)
        ax.set_xlabel("Range gate [\\#]")
        if next(first_it):
            ax.set_ylabel("Normalised power")
        ax.set_xticks(np.arange(0, 501, 100))
        legend_fontsize = 6
        ax.legend(
            **{**default_legend_kw, **{"fontsize": legend_fontsize}},
            title=f'S6-MF cycle {int(file_l1b.parent.stem.split("_")[13])}, pass {int(file_l1b.parent.stem.split("_")[14])},\n'
            f'lat={np.degrees(l1b["lat_rad"]):.3f}, lon={np.degrees(l1b["lon_rad"]):.3f},\n'
            f'dist-to-coast={l1b["dist2coast"]:.0f} km',
            title_fontsize=legend_fontsize,
        )
        ax.grid()

        # plot_retrack_result(l1b, l2, res_fit, model_params=model_params, wf_sets=wf_sets, model_sets=sr.model_sets, sensor_sets=sensor_sets,
        # retrack_sets=retrack_sets, ax=ax, show_l2_model=False,
        # show_second_halve=l1b_src_type == L1bSourceType.GPOD)

        # fn = ncfile_l1bs.name if ncfile_l1bs is not None else ''
        # fig.subplots_adjust(top=0.65)
        # fig.subplots_adjust(top=0.77)
        # fig.show()


# generate plots
gen_multilooked_vary_swh()
plot_single_retrackings()

for k, v in fig_list.items():
    v.show()

# export figures
export_in_jpg = []
for format in [
    "pgf",
    "pdf",
]:
    for figname, fig in fig_list.items():
        if figname in export_in_jpg:
            _f = "jpg"
        else:
            _f = format

        try:
            fig.savefig(THESIS_BASEDIR / f"{str(figname)}.{_f}", bbox_inches="tight")
            # fig.savefig(fig_export_dir / f'{figname}.{_f}')
        except Exception as e:
            print(f"error on saving {figname}: {e}")



import logging
from pathlib import Path

from pysamosa.common_types import RetrackerBaseType, L1bSourceType
from pysamosa.data_access import data_vars_s6
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset


if __name__ == "__main__":
    nc_src_base_path = Path(
        "/nfs/DGFI145/A/original_data/sentinel6a/JASON_CS_S6A_L1B_ALT_HR_NTC_F_V2/"
    )
    # nc_dest_path = Path('/nfs/DGFI145/C/work_flo/s6jtex_coral/')
    run_name = "coral_retrack"
    nc_dest_path = nc_src_base_path.parent

    # select files
    l1b_files = [f for f in sorted(nc_src_base_path.rglob("*.nc"))]

    l1b_src_type = (
        L1bSourceType.EUM_S6_F04
        if "f04" in str(l1b_files[0]).lower()
        else L1bSourceType.EUM_S6_F06
    )
    # rbt = RetrackerBaseType.SAMPLUS
    rbt = RetrackerBaseType.SAM
    pres = SettingsPreset.CORALv2
    # pres = SettingsPreset.NONE
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=rbt, settings_preset=pres, l1b_src_type=l1b_src_type
    )

    rp_sets.nc_dest_dir = nc_dest_path / run_name
    rp_sets.n_offset = 0
    rp_sets.n_inds = 0
    rp_sets.n_procs = 6
    rp_sets.skip_if_exists = False

    additional_nc_attrs = {
        "L1B source type": l1b_src_type.value.upper(),
        "Retracker basetype": rbt.value.upper(),
        "Retracker preset": pres.value.upper(),
    }

    rp = RetrackerProcessor(
        l1b_source=l1b_files,
        l1b_data_vars=data_vars_s6["l1b"],
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        wf_sets=wf_sets,
        sensor_sets=sensor_sets,
        nc_attrs_kw=additional_nc_attrs,
        log_level=logging.DEBUG,  # comment in to show debug messages
    )
    rp.process()



import logging
from pathlib import Path
import re
from datetime import datetime
import numpy as np
import pandas as pd

from pysamosa.common_types import RetrackerBaseType, L1bSourceType
from pysamosa.data_access import data_vars_s6
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset


def convert_dt(dt_str):
    return np.datetime64(datetime.strptime(dt_str, "%Y%m%dT%H%M%S"))


if __name__ == "__main__":
    nc_src_base_path = Path(
        "/nfs/DGFI145/A/original_data/sentinel6a/JASON_CS_S6A_L1B_ALT_HR_NTC_F_V2/"
    )
    run_name = "s6jtex_coral_reprocessed"
    nc_dest_path = Path("/nfs/DGFI145/C/work_flo/s6jtex_coral/")

    # cycle_pass_pattern = '_013_\d{3}_\d{3}_'
    cycle_pass_pattern = "_0(1[3-9]|2[0-2])_\\d{3}_\\d{3}_"
    # cycle_pass_pattern = '_022_\d{3}_\d{3}_'
    # cycle_pass_pattern = '_017_013_\d{3}_'

    # select files
    l1b_files = [
        f
        for f in sorted(nc_src_base_path.rglob("*.nc"))
        if bool(re.search(cycle_pass_pattern, str(f)))
    ]

    l1b_src_type = (
        L1bSourceType.EUM_S6_F04
        if "f04" in str(l1b_files[0]).lower()
        else L1bSourceType.EUM_S6_F06
    )
    rbt = RetrackerBaseType.SAM
    pres = SettingsPreset.CORALv2
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=rbt, settings_preset=pres, l1b_src_type=l1b_src_type
    )

    rp_sets.nc_dest_dir = nc_dest_path / run_name
    rp_sets.n_offset = 0
    rp_sets.n_inds = 0
    rp_sets.n_procs = 40
    rp_sets.skip_if_exists = True

    cycles = [int(f.name.split("_")[13]) for f in l1b_files]
    passes = [int(f.name.split("_")[14]) for f in l1b_files]
    start_dates = [convert_dt(f.name.split("_")[10]) for f in l1b_files]
    end_dates = [convert_dt(f.name.split("_")[11]) for f in l1b_files]

    df_l1b_files = pd.DataFrame(
        {
            "file": l1b_files,
            "cycle": cycles,
            "ppass": passes,
            "start_date": start_dates,
            "end_date": end_dates,
        }
    )
    df_l1b_files = df_l1b_files.drop_duplicates(["cycle", "ppass"])

    additional_nc_attrs = {
        "L1B source type": l1b_src_type.value.upper(),
        "Retracker basetype": rbt.value.upper(),
        "Retracker preset": pres.value.upper(),
    }

    rp = RetrackerProcessor(
        l1b_source=df_l1b_files.file.values,
        l1b_data_vars=data_vars_s6["l1b"],
        rp_sets=rp_sets,
        retrack_sets=retrack_sets,
        fitting_sets=fitting_sets,
        wf_sets=wf_sets,
        sensor_sets=sensor_sets,
        nc_attrs_kw=additional_nc_attrs,
        # log_level=logging.DEBUG,  #comment in to show
        # debug messages
    )
    rp.process()

    logging.shutdown()



import numpy as np
from datetime import datetime
import logging
from pathlib import Path
import re
import pandas as pd

from pysamosa.common_types import RetrackerBaseType, L1bSourceType, ProcMode
from pysamosa.data_access import data_vars_dart
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset

from pysamosa import simple_logger

simple_logger.set_root_logger(log_level=logging.INFO)


def convert_dt(dt_str):
    return np.datetime64(datetime.strptime(dt_str, "%Y%m%dT%H%M%S"))


TUDTUM_PATH = Path.home() / "TUDTUM"
COASTAL_FFSAR_PATH = Path("/lfs/DGFI24/bigdata/coastal_ffsar/")
# COASTAL_FFSAR_PATH = Path('/nfs/DGFI145/C/work_flo/coastal_ffsar/')

proc_configs = [
    # ffsar
    ((20, 20), "ffsar_dart", ProcMode.FFSAR, 2.1, data_vars_dart["l1b_ffsar"]),
    ((60, 20), "ffsar_dart", ProcMode.FFSAR, 2.1, data_vars_dart["l1b_ffsar"]),
    # ((100,20), 'ffsar_dart', ProcMode.FFSAR, 2.1, data_vars_dart['l1b_ffsar']),
    ((140, 20), "ffsar_dart", ProcMode.FFSAR, 2.1, data_vars_dart["l1b_ffsar"]),
    # ufsar
    ((20, 20), "ufsar_dart", ProcMode.UFSAR, 2.4, data_vars_dart["l1b"]),
    # ((60,20), 'ufsar_dart', ProcMode.UFSAR, 2.4, data_vars_dart['l1b']),
    # ((100,20), 'ufsar_dart', ProcMode.UFSAR, 2.4, data_vars_dart['l1b']),
    # ufsar-eum
    # ((20,20), 'ufsar_eum', ProcMode.UFSAR, 2.4, data_vars_eumetsat_s6['l1b']),
]

skip_if_exists = False
reverse_l1b_file_list = False

if __name__ == "__main__":
    # cycle_pass_pattern = '0(0[5-9]|[1-3][0-9]|4[0-2])_(120)'
    cycle_pass_pattern = "0(0[5-9]|[1-3][0-9]|4[0-2])_(213|018|196|120|044)"
    # cycle_pass_pattern = '0(0[5-9]|[1-3][0-9]|4[0-2])_(213)'
    # cycle_pass_pattern = '0(12|[2-3][0-9]|4[0-2])_(120)'
    # cycle_pass_pattern = '042_(213|018|196|120)'
    # cycle_pass_pattern = '026_120'

    for l1b_l2_posting_rate, procname, mode, T, data_vars in proc_configs:
        l1b_l2_path_append = f'_{procname.split("_")[-1]}' if "opt" in procname else ""
        l1b_posting_rate, l2_posting_rate = l1b_l2_posting_rate
        l1b_src_path_dart = (
            COASTAL_FFSAR_PATH
            / "processed"
            / "l1b"
            / f"T_{str(T).replace('.','-')}_posting_rate_{l1b_posting_rate}"
        )
        # l1b_src_path_eum = Path('/nfs/DGFI145/A/original_data/sentinel6a/JASON_CS_S6A_L1B_ALT_HR_NTC_F/')
        l1b_src_path_eum = Path(
            "/lfs/DGFI24/bigdata/coastal_ffsar/orig_data/f06/P4_1B_HR_____/"
        )

        # l2_destpath = COASTAL_FFSAR_PATH / f'coral_processed_L2_{l1b_posting_rate}_{l2_posting_rate}'
        l2_destpath = (
            COASTAL_FFSAR_PATH
            / "processed"
            / "l2"
            / f"{l1b_posting_rate}_{l2_posting_rate}"
        )

        if "dart" in procname and l1b_l2_path_append:
            data_vars["wf"] = f"power_waveform{l1b_l2_path_append}"

        l1b_files_dart = [
            f
            for f in sorted(l1b_src_path_dart.rglob("*.nc"))
            if "1B" in str(f) and bool(re.search(cycle_pass_pattern, str(f)))
        ]
        l1b_files_eum = [
            f
            for f in sorted(l1b_src_path_eum.rglob("*.nc"))
            if "1B" in str(f) and bool(re.search(cycle_pass_pattern, str(f)))
        ]
        l1b_files = l1b_files_dart if "dart" in procname else l1b_files_eum

        logging.info(
            f"Processing: l1b_l2_posting_rate: {l1b_l2_posting_rate}, len(l1b_files):{len(l1b_files)}, l1b_path:{l1b_src_path_dart} , l2_path:{l2_destpath}"
        )

        cycles = [int(f.name.split("_")[13]) for f in l1b_files_dart]
        passes = [int(f.name.split("_")[14]) for f in l1b_files_dart]
        start_dates = [convert_dt(f.name.split("_")[10]) for f in l1b_files_dart]
        end_dates = [convert_dt(f.name.split("_")[11]) for f in l1b_files_dart]
        bbox = [
            (float(f.stem.split("_")[21]), float(f.stem.split("_")[22]), 0, 360)
            for f in l1b_files_dart
        ]

        df_dart = pd.DataFrame(
            {
                "file": l1b_files_dart,
                "cycle": cycles,
                "ppass": passes,
                "start_date": start_dates,
                "end_date": end_dates,
                "dom": bbox,
            }
        )

        if reverse_l1b_file_list:
            df_dart = df_dart.iloc[::-1]

        is_ffsar = mode == ProcMode.FFSAR

        if is_ffsar:
            l1b_src_type = (
                L1bSourceType.EUM_S6_F04_FFSAR
                if "f04" in str(l1b_files_dart[0]).lower()
                else L1bSourceType.EUM_S6_F06_FFSAR
            )
        else:
            l1b_src_type = (
                L1bSourceType.EUM_S6_F04
                if "f04" in str(l1b_files_dart[0]).lower()
                else L1bSourceType.EUM_S6_F06
            )

        rbt, pres = RetrackerBaseType.SAM, SettingsPreset.CORALv2
        (
            rp_sets,
            retrack_sets,
            fitting_sets,
            wf_sets,
            sensor_sets,
        ) = get_default_base_settings(
            retracker_basetype=rbt, settings_preset=pres, l1b_src_type=l1b_src_type
        )

        rp_sets.nc_dest_dir = l2_destpath / procname
        rp_sets.n_offset = 0
        rp_sets.n_inds = 0
        rp_sets.n_procs = 20
        rp_sets.skip_if_exists = skip_if_exists
        rp_sets.reduce_l2_factor = round(l1b_posting_rate / l2_posting_rate)
        rp_sets.dynamic_fg_epoch_n_adjacent_meas *= rp_sets.reduce_l2_factor

        additional_nc_attrs = {
            "L1B source type": l1b_src_type.value.upper(),
            "Retracker basetype": rbt.value.upper(),
            "Retracker preset": pres.value.upper(),
        }

        rp = RetrackerProcessor(
            l1b_source=l1b_files,
            l1b_data_vars=data_vars,
            bbox=bbox,
            rp_sets=rp_sets,
            retrack_sets=retrack_sets,
            fitting_sets=fitting_sets,
            wf_sets=wf_sets,
            sensor_sets=sensor_sets,
            nc_attrs_kw=additional_nc_attrs,
            # log_level=logging.DEBUG,  #comment in to show
            # debug messages
        )
        rp.process()

    logging.shutdown()



from pysamosa.retracker import oversample_wf
import matplotlib.pyplot as plt
import numpy as np

from pysamosa.common_types import (
    RetrackerBaseType,
    WaveformSettings,
    SensorSettings,
    L1bSourceType,
    SensorType,
)
from pysamosa.model import SamosaModel, ModelParameter, ModelSettings


def test_setting_model_parameter_obj():
    mp = ModelParameter(ascending=True)
    print(mp)
    assert mp


def test_waveform_single_look():
    models_sets = ModelSettings.get_default_sets(
        st=SensorType.S3, retracker_basetype=RetrackerBaseType.SAM
    )
    sm = SamosaModel(
        sensor_sets=SensorSettings(), wf_sets=WaveformSettings(), model_sets=models_sets
    )

    # swh = np.arange(11) + 0.5
    swh = [0.25, 0.75, 1.0, 2.0, 5, 7, 9, 11]
    P_u = 1
    t_0 = -16

    for s in swh:
        wf = sm.get_waveform_singlelook(
            Pu=P_u, Hs=s, t0_ns=t_0, fa_Hz=0.0, model_params=ModelParameter()
        )
        # wf = wf + np.random.randn(len(wf)) * 0.01 * np.nanmax(wf)  # NOISE ADDITION (not in the SAMOSA documentation )
        # plt.plot(wf / np.nanmax(wf), label=f'swh={s}', linewidth=0.7)
        plt.plot(wf, label=f"swh={s}", linewidth=0.7)

    plt.legend()
    plt.grid()
    plt.title("SAMOSA single-look model")
    plt.ylabel("normalised power")
    plt.xlabel("# range bin")
    plt.show()


def test_multilooked_oversample():
    plt.figure(dpi=600)

    for oversampling_factor, m in zip([1, 4], ["x", None]):
        wf_sets = WaveformSettings.get_default_src_type(
            L1bSourceType.EUM_S3, internal_oversampling_factor=oversampling_factor
        )
        model_sets = ModelSettings(retracker_basetype=RetrackerBaseType.SAM)
        sm = SamosaModel(
            model_sets=model_sets, wf_sets=wf_sets, sensor_sets=SensorSettings()
        )

        swh = 3.0
        Pu = 1.0
        t_0_ns = 20
        nu = 0

        epoch_ref_gate = 64
        model_params = ModelParameter(epoch_ref_gate=epoch_ref_gate)
        wf = sm.get_waveform_multilook(
            Pu=Pu, Hs=swh, t0_ns=t_0_ns, nu=nu, model_params=model_params
        )

        # calculate interpolated x_vector to map all models on the same x-axis
        # scale
        x_interp = np.append(
            np.arange(0, 127 + 1 / oversampling_factor, 1 / oversampling_factor),
            np.arange(127 + 1 / oversampling_factor, 128, 1 / oversampling_factor),
        )

        plt.plot(
            x_interp,
            wf,
            linewidth=0.5,
            marker=m,
            markersize=0.8,
            label=f"wf_model Np={128*oversampling_factor}",
        )
        # plt.axvline(epoch_ref_gate, label=f'epoch_ref_gate (of={oversampling_factor})')

        assert not any(np.isnan(wf))

    plt.legend(fontsize=8)
    plt.xlabel("# range bin")
    plt.ylabel("normalised power")
    plt.grid()
    plt.show()


def test_different_model_len():
    plt.figure(dpi=600)

    oversampling_factor = 4

    for model_len, m in zip([509, 512, 500], ["x", None, "o"]):
        # for model_len, m in zip([512, 509],['x', None]):
        wf_sets = WaveformSettings.get_default_src_type(
            L1bSourceType.EUM_S3, zp_oversampling_factor=oversampling_factor
        )
        wf_sets.np = model_len
        model_sets = ModelSettings(retracker_basetype=RetrackerBaseType.SAM)
        sm = SamosaModel(
            model_sets=model_sets, wf_sets=wf_sets, sensor_sets=SensorSettings()
        )

        swh = 3.0
        Pu = 1.0
        t_0_ns = 20
        nu = 0

        epoch_ref_gate = 64
        model_params = ModelParameter(
            epoch_ref_gate=epoch_ref_gate * oversampling_factor
        )
        wf = sm.get_waveform_multilook(
            Pu=Pu, Hs=swh, t0_ns=t_0_ns, nu=nu, model_params=model_params
        )
        # wf = sm.get_waveform_singlelook(Pu=Pu, Hs=swh, t0_ns=t_0_ns, nu=nu, model_params=model_params, fa_Hz=0)

        x_gates = np.arange(0, model_len)

        plt.plot(
            x_gates,
            wf,
            linewidth=0.5,
            marker=m,
            markersize=1,
            label=f"wf_model_len={model_len}",
        )
        # plt.axvline(epoch_ref_gate, label=f'epoch_ref_gate (of={oversampling_factor})')

    plt.legend(fontsize=8)
    plt.xlabel("# range bin")
    plt.ylabel("normalised power")
    plt.grid()
    plt.show()


def test_oversampled_model():
    plt.figure(dpi=600)

    swh = 3.0
    Pu = 1.0
    t_0_ns = 20
    nu = 0
    epoch_ref_gate = 64

    internal_oversampling_factor = 4

    # oversampling=1 and oversample by internal_oversampling_Factor
    sm_of1 = SamosaModel(
        model_sets=ModelSettings(retracker_basetype=RetrackerBaseType.SAM),
        wf_sets=WaveformSettings.get_default_src_type(
            L1bSourceType.EUM_S3, internal_oversampling_factor=1
        ),
        sensor_sets=SensorSettings(),
    )
    model_params = ModelParameter(epoch_ref_gate=epoch_ref_gate)
    wf_of1 = sm_of1.get_waveform_multilook(
        Pu=Pu, Hs=swh, t0_ns=t_0_ns, nu=nu, model_params=model_params
    )

    wf_of1 = oversample_wf(wf_of1, internal_oversampling_factor)

    # oversampling=1 and oversample by internal_oversampling_Factor
    sm_of4 = SamosaModel(
        model_sets=ModelSettings(retracker_basetype=RetrackerBaseType.SAM),
        wf_sets=WaveformSettings.get_default_src_type(
            L1bSourceType.EUM_S3,
            internal_oversampling_factor=internal_oversampling_factor,
            # np=509,
        ),
        sensor_sets=SensorSettings(),
    )
    # model_params = ModelParameter(epoch_ref_gate=epoch_ref_gate * internal_oversampling_factor)
    model_params = ModelParameter(epoch_ref_gate=epoch_ref_gate)
    wf_of4 = sm_of4.get_waveform_multilook(
        Pu=Pu, Hs=swh, t0_ns=t_0_ns, nu=nu, model_params=model_params
    )

    # wf_len = 128 * internal_oversampling_factor
    # int_oversampling_fac = sm_of4.wf_sets.internal_oversampling_factor
    # wf_len_interp = wf_len - (int_oversampling_fac-1)
    #
    # interpolator = Akima1DInterpolator(np.arange(wf_len), wf_of4)
    # # x_interp = np.arange(0, wf_len - 1 + 1/int_oversampling_fac, 1/int_oversampling_fac)
    # # x_interp = np.linspace(0, wf_len_interp, wf_len_interp + (int_oversampling_fac - 1))
    # # x_interp = np.arange(0, wf_len_interp)
    # x_interp = np.linspace(0, wf_len - 1, wf_len_interp)
    # # x_interp_app = np.append(x_interp, np.arange(127 + 1/int_oversampling_fac, 128, 1 / int_oversampling_fac))
    #
    # wf_oversampled = np.full(wf_len, np.nan)
    # wf_oversampled[:-(int_oversampling_fac-1)] = interpolator(x_interp)
    # wf_oversampled[-(int_oversampling_fac-1):] = wf_oversampled[-(int_oversampling_fac)]  #replace last (int_oversampling_factor - 1) values by the last valid ones
    # # wf_oversampled = interpolator(x_interp)
    # wf_of4 = wf_oversampled

    plt.plot(
        wf_of1,
        linewidth=0.3,
        marker="x",
        markersize=0.1,
        label=f"model.internal_of={1} + oversample",
    )
    plt.plot(
        wf_of4,
        linewidth=0.3,
        marker="o",
        markersize=0.3,
        label=f"model.internal_of={internal_oversampling_factor}",
    )
    # plt.axvline(epoch_ref_gate, label=f'epoch_ref_gate (of={oversampling_factor})')

    plt.legend(fontsize=8)
    plt.xlabel("# range bin")
    plt.ylabel("normalised power")
    plt.grid()
    plt.show()


def test_multilooked_vary_swh():
    oversampling_factor = 1
    wf_sets = WaveformSettings.get_default_src_type(
        L1bSourceType.EUM_S3, zp_oversampling_factor=oversampling_factor
    )
    model_sets = ModelSettings(retracker_basetype=RetrackerBaseType.SAM)
    sm = SamosaModel(
        model_sets=model_sets, wf_sets=wf_sets, sensor_sets=SensorSettings()
    )

    swh = [0.25, 0.75, 1.0, 2.0, 5, 7, 9, 11]
    Pu = 1.0
    t_0 = 20
    nu = 0

    model_params = ModelParameter(epoch_ref_gate=64)
    for s in swh:
        wf = sm.get_waveform_multilook(
            Pu=Pu, Hs=s, t0_ns=t_0, nu=nu, model_params=model_params
        )
        # wf = wf + np.random.randn(len(wf)) * 0.01 * np.nanmax(wf)  # NOISE
        # ADDITION (not in the SAMOSA documentation )
        plt.plot(wf, label=f"swh={s}m", linewidth=1)

    plt.legend()
    plt.xlabel("# range bin")
    plt.ylabel("normalised power")
    plt.grid()
    plt.show()


def test_multilooked_vary_Pu():
    oversampling_factor = 1
    wf_sets = WaveformSettings.get_default_src_type(
        L1bSourceType.EUM_S3, zp_oversampling_factor=oversampling_factor
    )
    model_sets = ModelSettings(retracker_basetype=RetrackerBaseType.SAM)
    sm = SamosaModel(
        model_sets=model_sets, wf_sets=wf_sets, sensor_sets=SensorSettings()
    )

    swh = 2.0
    Pu = np.arange(0, 5, 1.0)
    t_0 = 20
    nu = 0

    model_params = ModelParameter(epoch_ref_gate=64)
    for p in Pu:
        wf = sm.get_waveform_multilook(
            Pu=p, Hs=swh, t0_ns=t_0, nu=nu, model_params=model_params
        )
        # wf = wf + np.random.randn(len(wf)) * 0.01 * np.nanmax(wf)  # NOISE
        # ADDITION (not in the SAMOSA documentation )
        plt.plot(wf, label=f"Pu={p}m", linewidth=1)

    plt.legend()
    plt.xlabel("# range bin")
    plt.ylabel("normalised power")
    plt.grid()
    plt.show()



import numpy as np
import pytest

from pysamosa.common_types import SENSOR_SETS_DEFAULT_S3
from pysamosa.retracker_helpers import (
    get_tracker_shift_n_samps,
    get_pointwise_product,
    shift_waveform_by_n,
    get_dynamic_first_guess_epochs,
    get_inds_sliding_window,
    gen_sliding_window_from_seq,
)


@pytest.mark.parametrize(
    "n_offset, n_inds",
    [
        (25800, 100),
    ],
)
def test_get_tracker_shift_n_samps(n_offset, n_inds, s3_eum_l1b):
    l1b_data = s3_eum_l1b(n_offset=n_offset, n_inds=n_inds, file_id="s3_0")

    n_shifts = get_tracker_shift_n_samps(
        tracker_range_m=l1b_data["tracker_range_m"][:20],
        alt_m=l1b_data["alt_m"][:20],
        t0_sampling_gate_ns=SENSOR_SETS_DEFAULT_S3.B_r_Hz,
    )

    assert len(n_shifts) == 20
    assert all(n_shifts < 3)


@pytest.mark.parametrize(
    "n_offset, n_inds",
    [
        (25800, 100),
    ],
)
def test_get_dynamic_first_guess_epochs_sam_coast(n_offset, n_inds, s3_eum_l1b):
    l1b_data = s3_eum_l1b(n_offset=n_offset, n_inds=n_inds, file_id="s3_0")

    fg_epoch = get_dynamic_first_guess_epochs(
        wfs=l1b_data["wf"],
        tracker_range=l1b_data["tracker_range_m"],
        alt_m=l1b_data["alt_m"],
        bu_bw_Hz=SENSOR_SETS_DEFAULT_S3.B_r_Hz,
        fg_epoch_adjacent_meas=40,
        # n_procs=None,
        n_procs=1,
    )

    assert len(fg_epoch) == l1b_data["wf"].shape[0]


@pytest.mark.parametrize(
    "n_offset, n_inds",
    [
        (25800, 1000),
    ],
)
def test_get_dynamic_first_guess_epochs_single_multcore(n_offset, n_inds, s3_eum_l1b):
    l1b_data = s3_eum_l1b(n_offset=n_offset, n_inds=n_inds, file_id="s3_0")

    fg_epoch = get_dynamic_first_guess_epochs(
        wfs=l1b_data["wf"],
        tracker_range=l1b_data["tracker_range_m"],
        alt_m=l1b_data["alt_m"],
        bu_bw_Hz=SENSOR_SETS_DEFAULT_S3.B_r_Hz,
        fg_epoch_adjacent_meas=40,
        n_procs=1,
    )

    fg_epoch_multcore = get_dynamic_first_guess_epochs(
        wfs=l1b_data["wf"],
        tracker_range=l1b_data["tracker_range_m"],
        alt_m=l1b_data["alt_m"],
        bu_bw_Hz=SENSOR_SETS_DEFAULT_S3.B_r_Hz,
        fg_epoch_adjacent_meas=40,
        n_procs=12,
    )

    assert len(fg_epoch) == l1b_data["wf"].shape[0]
    assert np.array_equal(fg_epoch, fg_epoch_multcore)


@pytest.mark.parametrize(
    "n_offset, n_inds",
    [
        (25800, 125),
    ],
)
def test_get_dynamic_first_guess_epochs_samplus(n_offset, n_inds, s3_eum_l1b):
    l1b_data = s3_eum_l1b(n_offset=n_offset, n_inds=n_inds, file_id="s3_0")

    fg_epoch = get_dynamic_first_guess_epochs(
        wfs=l1b_data["wf"],
        tracker_range=l1b_data["tracker_range_m"],
        alt_m=l1b_data["alt_m"],
        bu_bw_Hz=SENSOR_SETS_DEFAULT_S3.B_r_Hz,
    )

    assert len(fg_epoch) == l1b_data["wf"].shape[0]
    assert np.isclose(np.mean(fg_epoch), 44, atol=2)


def test_get_inds_sliding_window():
    exp_res = [
        ([0, 1, 2], 0),
        (
            [0, 1, 2],
            1,
        ),
        ([1, 2, 3], 1),
        ([2, 3, 4], 1),
        ([3, 4, 5], 1),
        ([3, 4, 5], 2),
    ]
    res = get_inds_sliding_window(n_total_len=6, n=3)
    assert res == exp_res

    exp_res = [
        ([0, 1, 2], 0),
        ([0, 1, 2], 1),
        ([1, 2, 3], 1),
        ([2, 3, 4], 1),
        ([2, 3, 4], 2),
    ]
    res = get_inds_sliding_window(n_total_len=5, n=3)
    assert res == exp_res

    exp_res = [
        ([0, 1, 2, 3], 0),
        ([0, 1, 2, 3], 1),
        ([0, 1, 2, 3], 2),
        ([1, 2, 3, 4], 2),
        ([1, 2, 3, 4], 3),
    ]
    res = get_inds_sliding_window(n_total_len=5, n=4)
    assert res == exp_res

    exp_res = [
        ([0, 1, 2], 0),
        ([0, 1, 2], 1),
        ([0, 1, 2], 2),
    ]
    res = get_inds_sliding_window(n_total_len=3, n=4)
    assert res == exp_res

    exp_res = [
        ([0, 1, 2, 3], 0),
        ([0, 1, 2, 3], 1),
        ([0, 1, 2, 3], 2),
        ([1, 2, 3, 4], 2),
        ([2, 3, 4, 5], 2),
        ([3, 4, 5, 6], 2),
        ([4, 5, 6, 7], 2),
        ([5, 6, 7, 8], 2),
        ([6, 7, 8, 9], 2),
        ([6, 7, 8, 9], 3),
    ]
    res = get_inds_sliding_window(n_total_len=10, n=4)
    assert res == exp_res

    exp_res = [
        ([0, 1, 2, 3, 4, 5], 0),
        ([0, 1, 2, 3, 4, 5], 1),
        ([0, 1, 2, 3, 4, 5], 2),
        ([0, 1, 2, 3, 4, 5], 3),
        ([1, 2, 3, 4, 5, 6], 3),
        ([2, 3, 4, 5, 6, 7], 3),
        ([3, 4, 5, 6, 7, 8], 3),
        ([4, 5, 6, 7, 8, 9], 3),
        ([4, 5, 6, 7, 8, 9], 4),
        ([4, 5, 6, 7, 8, 9], 5),
    ]
    res = get_inds_sliding_window(n_total_len=10, n=6)
    assert res == exp_res


def test_genslidingwindow():
    x = np.array([1] * 2 + [2] * 4 + [3] * 2 + [7])
    n = 4

    exp = [
        [1, 1, 2, 2, 2],
        [1, 1, 2, 2, 2],
        [1, 1, 2, 2, 2],
        [1, 2, 2, 2, 2],
        [2, 2, 2, 2, 3],
        [2, 2, 2, 3, 3],
        [2, 2, 3, 3, 7],
        [2, 2, 3, 3, 7],
        [2, 2, 3, 3, 7],
    ]

    exp_center = np.array([0, 1, 2, 2, 2, 2, 2, 3, 4])

    for i, (wnd, center_sample) in enumerate(gen_sliding_window_from_seq(x, n)):
        assert list(wnd) == exp[i]
        assert center_sample == exp_center[i]


def test_shift_waveform_by_n():
    X = np.tile(np.arange(10), 4).reshape(4, 10)

    res = shift_waveform_by_n(X, (0, 2, 3, -2))
    assert np.array_equal(
        res,
        np.array(
            [
                [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                [0, 0, 0, 1, 2, 3, 4, 5, 6, 7],
                [0, 0, 0, 0, 1, 2, 3, 4, 5, 6],
                [2, 3, 4, 5, 6, 7, 8, 9, 0, 0],
            ]
        ),
    )


def test_get_pointwise_product():
    X = np.tile(np.arange(10), 4).reshape(4, 10)

    assert all(
        get_pointwise_product(X)
        == np.array([0, 4, 16, 36, 64, 100, 144, 196, 256, 324])
    )
    assert all(get_pointwise_product(X, 1) == np.array([285, 285, 285, 285]))



from pathlib import Path


LFSDATA_DIR = Path(__file__).parent.parent / ".data"

S3_DATA_DIR = LFSDATA_DIR / "s3"
S6_DATA_DIR = LFSDATA_DIR / "s6"
FFSAR_DATA_DIR = LFSDATA_DIR / "s6" / "ffsar"
CS_DATA_DIR = LFSDATA_DIR / "cs"



import numpy as np
from numpy.random import default_rng

from pysamosa import simple_logger
from pysamosa.common_types import (
    ModelSettings,
    SensorSettings,
    RetrackerBaseType,
)
from pysamosa.conf_params import CONST_C
from pysamosa.data_access import get_model_param_obj_from_l1b_data
from pysamosa.model import SamosaModel

# fixed values taken from ind 42660, file
# RES_S3A_SR_1_SRA_A__20200108T101712_20200108T110742_20200202T201001_3029_053_279_GPOD_SAR_O_NT_003
l1b_data_single_template = {
    "wf": None,
    "lat_rad": -0.5918622555846702,
    "lon_rad": -0.3281458093122721,
    "alt_m": 815770.4278019192,
    "h_rate_m_per_s": 24.03475190775509,
    "Vs_m_per_s": 7534.799030740116,
    "ksix_rad": 2.094395102393196e-05,
    "ksiy_rad": 3.490658503988659e-06,
    "epoch_ref_gate": 64.0,
    "beam_ang_stack_rad": np.array([]),  # take ideal angles
    "dist2coast": 668,
}


class L1bSimulator:
    def __init__(
        self,
        *,
        model_sets: ModelSettings,
        swh,
        Pu=1.0,
        sensor_sets=None,
        wf_sets,
        add_thermal_speckle_noise=True,
        add_interference=False
    ):
        self.retracker_basetype = model_sets.retracker_basetype
        self.l1b_data_single = l1b_data_single_template

        self.model_sets = model_sets
        self.model_params = get_model_param_obj_from_l1b_data(
            l1b_data_single_template, 0
        )
        self.sensor_sets = sensor_sets if sensor_sets is not None else SensorSettings()
        self.wf_sets = wf_sets
        self.wf_len = self.wf_sets.np
        self.oversampling_factor = self.wf_sets.zp_oversampling_factor

        self.swh = swh
        self.dtau = 1 / (self.sensor_sets.B_r_Hz * self.wf_sets.zp_oversampling_factor)
        epoch_refgate = self.l1b_data_single["epoch_ref_gate"]
        retrack_point_gates = (
            38 if self.retracker_basetype == RetrackerBaseType.SAM else 310
        )
        self.epoch_ns = ((retrack_point_gates - epoch_refgate) * self.dtau) * 1e9
        self.Pu = Pu

        self.sam_model = SamosaModel(
            model_sets=self.model_sets, sensor_sets=self.sensor_sets, wf_sets=wf_sets
        )

        # noise
        self.rng = default_rng(42)  # generate noise generator object

        self.add_interference = add_interference
        self.add_thermal_speckle_noise = add_thermal_speckle_noise

        simple_logger.set_root_logger()

    def get_l2_data_single(self):
        return {
            "time": None,
            "lat_rad": None,
            "lon_rad": None,
            "swh": self.swh,
            "swh_qual": False,
            "Pu_W": self.Pu,
            "epoch_ns": self.epoch_ns,
            "range": (float(self.epoch_ns)) * 1e-9 * (CONST_C / 2),
            "tracker_range_m": 0.0,
            "alt_m": l1b_data_single_template["alt_m"],
            "n_iter": 0,
            "misfit": 0.0,
        }

    def get_awgn(self):
        # generate additive white gaussian noise (AWGN)
        mean, variance = 5e-3, 5e-4
        return mean + variance * self.rng.standard_normal(self.wf_len)

    def get_multi_noise(self):
        k = 200
        # k = 500
        # mult_noise = self.rng.gamma(shape=k, scale=0.5 * 1/k, size=self.wf_len)
        mult_noise = self.rng.gamma(shape=k, scale=0.4, size=self.wf_len)
        return mult_noise

    def get_interference_inds(self, wf):
        # interference = np.zeros(self.wf_len)

        wf_maxgate = np.argmax(wf)
        diff_start_gate_from_max = 15 * self.oversampling_factor

        rand_start_gate = (
            int(self.rng.uniform(-5, 5)) * self.oversampling_factor
        )  # randomised start_gate
        start_gate = wf_maxgate + diff_start_gate_from_max + rand_start_gate
        # randomised width of interference
        rand_dist_width = int(self.rng.uniform(10, 20)) * self.oversampling_factor

        # interference[start_gate:start_gate+rand_dist_width] = 1.0
        interference_mask = np.zeros(self.wf_len, dtype=bool)
        interference_mask[start_gate : start_gate + rand_dist_width] = True

        return interference_mask

    def __iter__(self):
        return self

    def __next__(self):
        wf = self.sam_model.get_waveform_multilook(
            Pu=self.Pu,
            Hs=self.swh,
            t0_ns=self.epoch_ns,
            nu=0,
            model_params=self.model_params,
        )
        wf = wf / np.max(wf)  # normalise waveform

        # add additive (thermal noise) and multiplicative (speckle) noise
        if self.add_thermal_speckle_noise:
            wf += self.get_awgn()
            wf *= self.get_multi_noise()

        wf = wf / np.max(wf)  # normalise waveform

        wf_argmax = int(np.argmax(wf))

        if self.add_interference:
            distorted_mask = self.get_interference_inds(wf=wf)
            # set distorted gates to a maximum waveform value of 1.0 (omit the
            # noise in these gates)
            wf[distorted_mask] = 1.0
            wf_argmax = int(np.argmax(wf[~distorted_mask]))

        return {
            **self.l1b_data_single,
            **{
                "wf": wf,
                "dynamic_fg_epoch": wf_argmax,
            },
        }



import pytest
import matplotlib.pyplot as plt

from pysamosa.common_types import (
    ModelSettings,
    RetrackerBaseType,
    WaveformSettings,
    L1bSourceType,
    SensorSettings,
    SensorType,
)
from pysamosa.l1b_simulator import L1bSimulator


@pytest.mark.parametrize(
    "swh",
    # np.arange(1, 10, 2),
    [2.0],
)
def test_gen_l1b_data_single(swh):
    n_realisations = 3
    rbt = RetrackerBaseType.SAM
    # rbt = RetrackerBaseType.SAMPLUS
    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3)
    sensor_sets = SensorSettings.get_default_sets(st=SensorType.S3)

    model_sets = ModelSettings.get_default_sets(
        st=sensor_sets.sensor_type,
        retracker_basetype=rbt,
        n_effective_looks=0,
        wf_sets=wf_sets,
    )

    l1b_sim = L1bSimulator(
        model_sets=model_sets, swh=swh, Pu=1.0, sensor_sets=sensor_sets, wf_sets=wf_sets
    )
    l1b_sim_it = iter(l1b_sim)

    for i in range(n_realisations):
        l1b_data_single = next(l1b_sim_it)
        plt.plot(l1b_data_single["wf"], linewidth=0.7)
        assert l1b_data_single["wf"] is not None

    plt.title(f"SWH={swh}m")
    plt.grid()
    plt.show()


@pytest.mark.parametrize(
    "swh",
    # np.arange(1, 10, 2),
    [2.0],
)
def test_gen_l1b_data_add_interference(swh):
    n_realisations = 3
    rbt = RetrackerBaseType.SAM
    # rbt = RetrackerBaseType.SAMPLUS
    wf_sets = WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3)
    sensor_sets = SensorSettings.get_default_sets(st=SensorType.S3)

    model_sets = ModelSettings.get_default_sets(
        st=sensor_sets.sensor_type,
        retracker_basetype=rbt,
        n_effective_looks=0,
        wf_sets=wf_sets,
    )

    l1b_sim = L1bSimulator(
        model_sets=model_sets,
        swh=swh,
        Pu=1.0,
        sensor_sets=SensorSettings(),
        add_interference=True,
        wf_sets=WaveformSettings.get_default_src_type(L1bSourceType.EUM_S3),
    )
    l1b_sim_it = iter(l1b_sim)

    for i in range(n_realisations):
        l1b_data_single = next(l1b_sim_it)
        plt.plot(l1b_data_single["wf"], linewidth=0.7)
        assert l1b_data_single["wf"] is not None

    plt.title(f"SWH={swh}m")
    plt.grid()
    plt.show()



from pathlib import Path
import pickle
import numpy as np
import xarray as xr

import matplotlib.pyplot as plt

from pysamosa.common_types import SensorType


def read_f(name):
    LUT_X = []
    LUT_Y = []
    with open(name) as fid:
        for f in fid:
            if f[0] == "#":
                continue
            line = f.split()
            LUT_X.append(float(line[0]))
            LUT_Y.append(float(line[1]))

    LUT_X = np.array(LUT_X)
    LUT_Y = np.array(LUT_Y)
    return LUT_X, LUT_Y


def read_lut_alpha_p(filename_lut_alpha_p):
    SWH_lut, alpha_p_lut = [], []

    # filename_lut_alpha_p = 'LUT_Alpha_P_CS-2.txt'  # CS-2
    # filename_lut_alpha_p = 'alphap_table_SEN3_09_Nov_2017.txt'  # S3A

    with open(filename_lut_alpha_p) as fid:
        for li in fid:
            if li[0] == "#":
                continue
            line = li.split()
            SWH_lut.append(float(line[0]))
            alpha_p_lut.append(float(line[1]))

    return SWH_lut, alpha_p_lut


# S3 LUTs
LUT_X0, LUT_Y0 = read_f("F0.txt")
LUT_X1, LUT_Y1 = read_f("F1.txt")

F0 = np.transpose(np.array([LUT_X0, LUT_Y0]))
F1 = np.transpose(np.array([LUT_X1, LUT_Y1]))

lut_alpha_p_s3 = np.transpose(
    np.array(read_lut_alpha_p("alphap_table_SEN3_09_Nov_2017.txt"))
)

luts_s3 = {"lut_alpha_p": lut_alpha_p_s3, "lut_F0": F0, "lut_F1": F1}

# CS LUTs
LUT_X0, LUT_Y0 = read_f("F0.txt")
LUT_X1, LUT_Y1 = read_f("F1.txt")

F0 = np.transpose(np.array([LUT_X0, LUT_Y0]))
F1 = np.transpose(np.array([LUT_X1, LUT_Y1]))

lut_alpha_p_cs = np.transpose(np.array(read_lut_alpha_p("LUT_Alpha_P_CS-2.txt")))

luts_cs = {"lut_alpha_p": lut_alpha_p_cs, "lut_F0": F0, "lut_F1": F1}

# S6 F04 LUTs


def read_s6_lut(lut_file):
    with xr.open_dataset(lut_file) as ds:
        F0 = np.transpose(
            np.array([ds.LUT_F0_X.values, ds.LUT_F0_Y.values], dtype="float64")
        )
        F1 = np.transpose(
            np.array([ds.LUT_F1_X.values, ds.LUT_F1_Y.values], dtype="float64")
        )

        alpha_p_x_app = np.hstack(
            [np.linspace(-1, 0, 100, endpoint=False), ds.AlphaP_X.values]
        )
        alpha_p_y_app = np.hstack(
            [np.repeat(ds.AlphaP_Y.values[0], 100), ds.AlphaP_Y.values]
        )
        lut_alpha_p = np.transpose(np.array([alpha_p_x_app, alpha_p_y_app]))

        return {"lut_alpha_p": lut_alpha_p, "lut_F0": F0, "lut_F1": F1}


luts_s6_f04 = read_s6_lut("AUX_RLUT_S6A_002.nc")
luts_s6_f06 = read_s6_lut("AUX_RLUT_S6A_003.nc")
luts_s6_f06_ff = read_s6_lut(
    "S6A_TEST_AUX_FLUT___00000000T000000_99999999T999999_0001.NC"
)

all_luts = {
    SensorType.CS.value: luts_cs,
    SensorType.S3.value: luts_s3,
    SensorType.S6_F04.value: luts_s6_f04,
    SensorType.S6_F06.value: luts_s6_f06,
    SensorType.S6_F06_FF.value: luts_s6_f06_ff,
}

destfile_pickle = Path.cwd().parent.parent / "pysamosa" / "luts_samosa.pickle"
with open(destfile_pickle, "wb") as handle:
    pickle.dump(all_luts, handle)

# plot LUTS
for mission, luts in all_luts.items():
    plt.plot(luts["lut_alpha_p"][:, 0], luts["lut_alpha_p"][:, 1], label=mission)

plt.legend()
plt.grid()
plt.ylabel("alpha_p")
plt.xlabel("SWH [m]")
plt.show()

print(f"LUTs successfully written to {destfile_pickle}. ")



import logging
from pathlib import Path
import re
import pandas as pd

from pysamosa.common_types import RetrackerBaseType, L1bSourceType
from pysamosa.data_access import data_vars_s6
from pysamosa.retracker_processor import RetrackerProcessor
from pysamosa.settings_manager import get_default_base_settings, SettingsPreset


if __name__ == "__main__":
    nc_dest_path = Path("/lfs/DGFI24/bigdata/s6jtex_raw2rmc/retracked/")

    run_name_basepath = [
        ("coral_raw", Path("/lfs/DGFI24/bigdata/s6jtex_raw2rmc/HR_RAW/1B/")),
        ("coral_rmc", Path("/lfs/DGFI24/bigdata/s6jtex_raw2rmc/HR_RMC/1B/")),
    ]

    l1b_src_type, rbt, pres = (
        L1bSourceType.EUM_S6_F04,
        RetrackerBaseType.SAM,
        SettingsPreset.CORALv2,
    )
    (
        rp_sets,
        retrack_sets,
        fitting_sets,
        wf_sets,
        sensor_sets,
    ) = get_default_base_settings(
        retracker_basetype=rbt, settings_preset=pres, l1b_src_type=l1b_src_type
    )

    for run_name, nc_src_base_path in run_name_basepath:
        rp_sets.nc_dest_dir = nc_dest_path / run_name
        rp_sets.n_offset = 0
        rp_sets.n_inds = 0
        rp_sets.n_procs = 4
        rp_sets.skip_if_exists = False

        # select files
        l1b_files = [
            f
            for f in sorted(nc_src_base_path.rglob("*.nc"))
            if bool(
                re.search(
                    "02(5|6|7)_(018|044|120|196|213)_\\d{3}_EUM", str(f.parent.name)
                )
            )
        ]
        # l1b_files = [f for f in sorted(nc_src_base_path.rglob('*.nc')) if bool(re.search('02(5|6|7)_(018|044)_\d{3}_EUM', str(f.parent.name)))]
        df_l1b_files = pd.DataFrame(
            {
                "file": l1b_files,
                "cycle": [int(f.parent.name.split("_")[13]) for f in l1b_files],
                "ppass": [int(f.parent.name.split("_")[14]) for f in l1b_files],
            }
        )
        bboxes_per_ppass = {
            18: (53.65, 53.88, 0, 360),
            44: (50.99, 51.406, 0, 360),
            120: (51.794, 52.05, 0, 360),
            196: (53.13, 53.375, 0, 360),
            213: (53.73, 53.95, 0, 360),
        }
        bboxes_all = 0

        additional_nc_attrs = {
            "L1B source type": l1b_src_type.value.upper(),
            "Retracker basetype": rbt.value.upper(),
            "Retracker preset": pres.value.upper(),
        }

        for p in df_l1b_files.ppass.unique():
            sel_files = df_l1b_files[df_l1b_files.ppass == p]

            rp = RetrackerProcessor(
                l1b_source=sel_files.file.to_list(),
                l1b_data_vars=data_vars_s6["l1b"],
                bbox=[bboxes_per_ppass[p] for i in range(len(sel_files))],
                rp_sets=rp_sets,
                retrack_sets=retrack_sets,
                fitting_sets=fitting_sets,
                wf_sets=wf_sets,
                sensor_sets=sensor_sets,
                nc_attrs_kw=additional_nc_attrs,
                # log_level=logging.DEBUG,  #comment in to
                # show debug messages
            )
            rp.process()

    logging.shutdown()



import os
from pathlib import Path

from tqdm import tqdm
import requests
import tarfile

import numpy as np
import xarray as xr
from scipy import ndimage


DEFAULT_DUMMY_LAT = 48.1411
DEFAULT_DUMMY_LON = 11.5777

DATA_DIR = Path(__file__).parent.parent / ".data"

download_files = {
    "d2c": {
        "url": "https://www.dropbox.com/s/1jhren60kmysr0w/dist2coast_1deg_merged.tar?dl=1",
        "size": 330857131,
    },
    "pysamosa_data": {
        "url": "https://www.dropbox.com/s/ygagj9gebnhsm5b/pysamosa_data.tar?dl=1",
        "size": 218451704,
    },
}


def download_dist2coast_nc():
    id = "d2c"
    url = download_files[id]["url"]
    fsize = download_files[id]["size"]
    fpath = DATA_DIR / Path(url).name.split("?")[0]
    print(fpath)

    # delete corrupt file (with wrong size!?)
    if fpath.exists() and fpath.stat().st_size != fsize:
        fpath.unlink()

    download_untar_file(url=url, dest_file=fpath, total_file_size=fsize)


def download_pysamosa_data():
    id = "pysamosa_data"
    url = download_files[id]["url"]
    fsize = download_files[id]["size"]
    fpath = DATA_DIR / Path(url).name.split("?")[0]

    # delete corrupt file (with wrong size!?)
    if fpath.exists() and fpath.stat().st_size != fsize:
        fpath.unlink()

    download_untar_file(url=url, dest_file=fpath, total_file_size=fsize)


def download_untar_file(url: str, dest_file: str, total_file_size: int = None):
    os.umask(000)
    dest_file.parent.mkdir(parents=True, exist_ok=True)

    with open(dest_file, "wb") as f:
        with tqdm(
            total=total_file_size,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
            desc=f"Downloading {dest_file}... (Total {total_file_size / (1024**3):.2f}GB)",
            ascii=True,
        ) as pbar:
            for chunk in requests.get(url, stream=True).iter_content(32 * 1024):
                f.write(chunk)
                pbar.update(len(chunk))

    # extract downloaded zip file
    with tarfile.open(dest_file) as tar:
        tar.extractall(path=dest_file.parent)


def get_dist_pacioos(latarr, lonarr, do_interp=True, pre_loaded_chunk_size=None):
    """
    Calculates distance to coast based on PacIOOS source.
    By now for ocean only.

    Online resource:
    - Ocean: http://www.pacioos.hawaii.edu/metadata/dist2coast_1deg_ocean.html
    - Land: http://www.pacioos.hawaii.edu/metadata/dist2coast_1deg_land.html

    :param latarr: latitude in decimal degrees, ndarray
    :param lonarr: longitude in decimal degrees, assumed to be within range (-180, 180] degrees, ndarray
    :param do_interp: if True, interpolation is active and distance to coast is calculated for coords outside 0.01-degree grid
    :return: distance to coast in km
    """
    id = "d2c"
    url = download_files[id]["url"]
    fsize = download_files[id]["size"]
    fpath = DATA_DIR / Path(url).name.split("?")[0]
    nc_file = fpath.parent / (fpath.stem + ".nc")

    # delete corrupt file (with wrong size!?)
    if fpath.exists() and fpath.stat().st_size != fsize:
        download_dist2coast_nc()

    try:
        _latarr = np.asarray(latarr)
        _lonarr = np.asarray(lonarr)
    except BaseException:
        raise ValueError("latarr, lonarr are not convertible to ndarrays. ")

    if _latarr.size != _lonarr.size:
        raise ValueError

    np.seterr(invalid="ignore")
    if ((-180 < _lonarr) & (_lonarr >= 180.0)).any():
        _lonarr[_lonarr >= 180.0] -= 360.0

    _step = 0.01
    _first_lat = 90.00
    _first_lon = -180.00

    with xr.open_dataset(
        nc_file, engine="h5netcdf", chunks=pre_loaded_chunk_size
    ) as ds:
        if do_interp:

            def func_lat_ind_float(_lat):
                return -(_lat - _first_lat) / _step

            def func_lon_ind_float(_lon):
                return (_lon - _first_lon) / _step

            res = ndimage.map_coordinates(
                ds.dist,
                [func_lat_ind_float(latarr), func_lon_ind_float(lonarr)],
                order=1,
            )
        else:
            # trunc = lambda arr, n_dec: np.trunc(arr * 10 ** n_dec) / 10 ** n_dec
            avoid_round_error = 0.001
            lat_ind = (
                np.round(-(_latarr - _first_lat) / _step + avoid_round_error)
            ).astype(int)
            lon_ind = (
                np.round((_lonarr - _first_lon) / _step + avoid_round_error)
            ).astype(int)

            res = ds.dist.isel(
                lat=xr.DataArray(lat_ind % ds.dist.shape[0]),
                lon=xr.DataArray(lon_ind % ds.dist.shape[1]),
            ).values

    # insert nans where there are in the original latarr, lonarr
    lat_nan_inds = np.isnan(_latarr)
    lon_nan_inds = np.isnan(_lonarr)
    if np.isnan([latarr, lonarr]).any():
        res[lat_nan_inds | lon_nan_inds] = np.nan

    return res
